# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: University of Washington <https://www.washington.edu>
# SPDX-FileContributor: 2021-24 Bradley M. Bell
# ----------------------------------------------------------------------------
'''
{xrst_begin map_shared}

Map at_cascade Shared Memory Names
##################################

Prototype
*********
{xrst_literal ,
   # BEGIN_DEF , # END_DEF
   # BEGIN_RETURN , # END_RETURN
}

Purpose
*******
The mac Darwin system has a limit on the length (in characters) of
shared memory names. This routine returns a mapping of the shared memory
name that is short enough to work on the current system.

{xrst_end map_shared}
'''
# BEGIN_DEF
# at_cascade.map_shared
import platform
import os
def map_shared(shared_name) :
   assert type(shared_name) == str
   # END_DEF
   mac_os  = platform.system() == 'Darwin'
   sandbox  = os.getcwd().endswith('at_cascade.git')
   if mac_os or sandbox :
      hash_code   = 0
      factor      = 0
      for ch in shared_name :
         factor    += 1
         hash_code += factor * ord(ch)
      mapped_name = str( hash_code )
   else :
      # This case is not tested by bin/check_all.sh
      mapped_name = shared_name
   # BEGIN_RETURN
   assert type(mapped_name) == str
   return mapped_name
   # END_RETURN



# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: University of Washington <https://www.washington.edu>
# SPDX-FileContributor: 2021-24 Bradley M. Bell
# ----------------------------------------------------------------------------
r'''
{xrst_begin create_job_table}
{xrst_spell
  bool
}

Table of Job Parent Child Relationships
#######################################

Prototype
*********
{xrst_literal ,
   # BEGIN_DEF, END_DEF
   # BEGIN_RETURN, END_RETURN
}

Summary
*******
This routine returns a list where each element corresponds to a job:

#. A job is a combination of a node and split reference value.
   For example, if the node is n0 and we are splitting on
   sex some possible jobs are n0.female, n0.male.

#. All of the jobs that have
   :ref:`create_job_table@job_table@prior_only` false,
   must be fit to fit all the jobs for the nodes in the *fit_goal_set*.

#. Each job has a *parent_job_id* for the job that needs to be fit before it,
   except for the start job which corresponds to the start node
   and start split reference id.
   The *prior_only* field is false for any job that is a parent; i.e.,
   all the parent jobs are fit.

#. Each job also has a list of which jobs need be run after it
   (to fit the *fit_goal_set* ).

#. If a job has *prior_only* true,
   it does not need to be fit for this *fit_goal_set*, but its priors should
   be created (when the corresponding parent job is fit)
   so it could be the start job for a different *fit_goal_set* .



all_node_database
*****************
is a python string specifying the location of the
:ref:`all_node_db-name`
relative to the current working directory.
This argument can't be ``None``.

node_table
**********
is a ``list`` of ``dict`` containing the node table for this cascade.
This argument can't be ``None``.

start_node_id
*************
This, together with *start_split_reference_id*
corresponds to a completed fit that we are starting from.
We assume that the priors for this fit have been created; see prior_only below.
The start node must be a descendant of the :ref:`glossary@root_node` .

start_split_reference_id
************************
This, together with *start_node_id*
corresponds to a completed fit that we are starting from.
Only jobs that depend on the start jobs completion will be included in
the job table.
This is ``None`` if and only if
:ref:`split_reference_table-name` is empty.

fit_goal_set
************
This is the a :ref:`glossary@fit_goal_set`.
In addition,
each such node must be the start node, or a descendant of the start node.

job_table
*********
The return value *job_table* is a ``list`` of ``dict`` :

job_id
======
We use *this_job_id* to denote the index of a row in the job_table list.
The value *job_table[job_id]* is a ``dict`` with the following keys:

job_name
========
This is a ``str`` containing the job name.
If the :ref:`split_reference_table-name` is empty,
*job_name* is equal to *node_name*
where *node_name* is the node name corresponding to *node_id*.
Otherwise, *job_name* is equal to
*node_name*\ ``.``\ *split_reference_name*
where *split_reference_name* is the split reference name corresponding to
*split_reference_id*.

prior_only
==========
If this ``bool`` is false,
this job must be run to fit all the nodes in *fit_goal_set* .
It will be false if this is the start job; i.e,
the start job must be fit to fit the nodes in *fit_goal_set*.

If *prior_only* is true,
*prior_only* cannot be true for the corresponding parent job.
The priors for this job will be created if the parent job succeeds,
but this job will not be run and it will not have any children.
These priors are intended to be used by a subsequent call to
:ref:`continue_cascade-name` where this job is the start job
( and *prior_only* is false because we have a different *fit_goal_set* ).

fit_node_id
===========
This is an ``int`` containing the node_id for the
:ref:`glossary@fit_node` for this *this_job_id*.

split_reference_id
==================
If the split_reference table is empty, this is ``None``.
Otherwise it is an ``int`` containing the
:ref:`split_reference_table@split_reference_id`
for this *this_job_id*; i.e. the splitting covariate has this reference
value.

parent_job_id
=============
This is an ``int`` containing the job_id corresponding to the parent job
which must be greater than the job_id for this row of the job table.
The parent job (and only the parent job)
must have completed before this job can be run.
This first row of the job table has *parent_job_id* equal to None; i.e.,
there is not parent for the first job.

start_child_job_id
==================
This is the job_id for the first job that can run as soon as this job
is completed. The start_child_job_id is always greater than the job_id
for the current row. The simplest way to run the jobs is in job table
order (not in parallel).

end_child_job_id
================
This is the job_id plus one for the last job that can run as soon as
this job is completed. If end_child_job_id is equal to start_child_job_id,
there are no jobs that require the results of this job.
Note that this job is the parent of each job between the start and end,


{xrst_end create_job_table}
'''
# -----------------------------------------------------------------------------
import dismod_at
import at_cascade
# -----------------------------------------------------------------------------
def ancestor_set(node_table, node_id) :
   result = { node_id }
   while node_table[node_id]['parent'] != None :
      node_id = node_table[node_id]['parent']
      result.add(node_id)
   return result
# -----------------------------------------------------------------------------
def get_child_job_table(
   job_id                     ,
   fit_node_id                ,
   refit_split                ,
   fit_split_reference_id     ,
   root_split_reference_id    ,
   split_reference_table      ,
   node_split_set             ,
   fit_children               ,
   prior_children             ,
   node_table                 ,
) :
   assert type(job_id) == int
   assert type(fit_node_id) == int
   assert type(refit_split) == bool
   assert type(fit_split_reference_id) ==int or fit_split_reference_id  == None
   assert type(root_split_reference_id)==int or root_split_reference_id == None
   assert type(node_split_set) == set
   assert type(fit_children) == list
   assert type(prior_children) == list
   if len(fit_children) > 0 :
      assert type( fit_children[0] ) == set
      assert type( prior_children[0] ) == set
   assert type(node_table) == list
   if len(node_table) > 0 :
      assert type( node_table[0] ) == dict
   #
   # already_split
   already_split = root_split_reference_id != fit_split_reference_id
   #
   # shift_reference_set
   if already_split or fit_node_id not in node_split_set :
      shift_reference_set = { fit_split_reference_id }
   else :
      shift_reference_set = set( range( len(split_reference_table) ) )
      shift_reference_set.remove( root_split_reference_id )
   #
   # fit_node_set, shift_node_set
   if fit_node_id in node_split_set and not already_split and refit_split :
      shift_node_set = { fit_node_id }
      fit_node_set   = { fit_node_id }
   else :
      shift_node_set = prior_children[ fit_node_id ]
      fit_node_set   = fit_children[ fit_node_id ]
   #
   # child_job_table
   child_job_table = list()
   for shift_split_reference_id in shift_reference_set :
      for shift_node_id in shift_node_set :
         #
         # prior_only
         prior_only = shift_node_id not in fit_node_set
         #
         # job_name
         job_name = node_table[shift_node_id]['node_name']
         if shift_split_reference_id != None :
            row  = split_reference_table[shift_split_reference_id]
            split_reference_name = row['split_reference_name']
            job_name             = f'{job_name}.{split_reference_name}'
         #
         # child_job_table
         row = {
            'job_name'           : job_name,
            'prior_only'         : prior_only,
            'fit_node_id'        : shift_node_id,
            'split_reference_id' : shift_split_reference_id,
            'parent_job_id'      : job_id,
         }
         child_job_table.append( row )
   #
   return child_job_table
# -----------------------------------------------------------------------------
# BEGIN_DEF
# at_cascade.create_job_table
def create_job_table(
   all_node_database                 ,
   node_table                        ,
   start_node_id                     ,
   fit_goal_set                      ,
   start_split_reference_id   = None ,
) :
   assert type(all_node_database) == str
   assert type(node_table) == list
   if len(node_table) > 0 :
      assert type( node_table[0] ) == dict
   assert type(start_node_id) == int
   assert type(start_split_reference_id) == int or \
      start_split_reference_id == None
   assert type(fit_goal_set) == set
   # END_DEF
   #
   # fit_goal_set
   temp = set()
   for node in fit_goal_set :
      if type(node) == str :
         node_id = at_cascade.table_name2id(node_table, 'node', node)
      else :
         assert type(node) == int
         node_id = node
      temp.add(node_id)
   fit_goal_set = temp
   #
   # all_table
   all_table = dict()
   connection = dismod_at.create_connection(
      all_node_database, new = False, readonly = True
   )
   tbl_list   =  [ 'option_all', 'split_reference', 'node_split', 'fit_goal' ]
   for name in tbl_list :
      all_table[name] = dismod_at.get_table_dict(connection, name)
   connection.close()
   #
   # prior_goal_set
   prior_goal_set = set()
   if len(all_table['fit_goal']) == 0 :
      prior_goal_set = set( range( len(node_table) ) )
   else :
      for row in all_table['fit_goal'] :
         prior_goal_set.add( row['node_id'] )
   #
   # fit_goal_ancestor
   fit_goal_ancestor = set()
   if len(all_table['fit_goal']) == 0 :
      fit_goal_ancestor = set( range( len(node_table) ) )
   else :
      for row in all_table['fit_goal'] :
         node_id = row['node_id']
         fit_goal_ancestor = \
            fit_goal_ancestor.union( ancestor_set(node_table, node_id ) )
   #
   for node_id in fit_goal_set :
      if start_node_id not in ancestor_set(node_table, node_id) :
         node_name       = node_table[node_id]['node_name']
         start_node_name = node_table[start_node_id]['node_name']
         msg  = f'create_job_table: node {node_name} is in fit_goal_set but\n'
         msg += 'it is not a descendant of the start node {start_node_name}'
         assert False, msg
      if node_id not in fit_goal_ancestor :
         node_name = node_table[node_id]['node_name']
         msg  = f'create_job_table: node {node_name} is in fit_goal_set\n'
         msg += 'but it is not in the fit_goal table or the ancestor of'
         msg += 'a node in the fit_goal table.'
         assert False, msg
   #
   # node_split_set
   node_split_set = set()
   for row in all_table['node_split'] :
      node_split_set.add( row['node_id'] )
   #
   # option_all_dict
   option_all_dict = dict()
   for row in all_table['option_all'] :
      option_all_dict[ row['option_name'] ] = row['option_value']
   #
   # refit_split
   if 'refit_split' in option_all_dict :
      refit_split = option_all_dict['refit_split']
      assert refit_split in [ 'true', 'false' ]
      refit_split = refit_split == 'true'
   else :
      refit_split = False
   #
   # root_node_name
   assert 'root_node_name' in option_all_dict
   root_node_name = option_all_dict['root_node_name']
   #
   # root_node_id
   root_node_id = at_cascade.table_name2id(node_table, 'node', root_node_name)
   #
   # fit_children
   fit_children = at_cascade.get_fit_children(
      root_node_id, fit_goal_set, node_table
   )
   #
   # prior_children
   prior_children = at_cascade.get_fit_children(
      root_node_id, prior_goal_set, node_table
   )
   #
   # root_split_reference_id
   if 'root_split_reference_name' in option_all_dict :
      root_split_reference_name = option_all_dict['root_split_reference_name']
      root_split_reference_id   = at_cascade.table_name2id(
         all_table['split_reference'],
         'split_reference',
         root_split_reference_name
      )
   else :
      root_split_reference_id   = None
      if refit_split :
         msg  = 'option_all_table: refit_split is true and '
         msg += ' root_split_reference_name does not appear'
         assert False, msg
   #
   # job_name
   job_name = node_table[start_node_id]['node_name']
   if start_split_reference_id != None :
      row       = all_table['split_reference'][start_split_reference_id]
      split_reference_name = row['split_reference_name']
      job_name             = f'{job_name}.{split_reference_name}'
   #
   # job_table
   job_table = [ {
      'job_name'           : job_name,
      'prior_only'         : False,
      'fit_node_id'        : start_node_id,
      'split_reference_id' : start_split_reference_id,
      'parent_job_id'      : None,
   } ]
   #
   # job_id
   job_id = 0
   #
   while job_id < len(job_table) :
      #
      # row, prior_only
      row                = job_table[job_id]
      prior_only         = row['prior_only']
      #
      if not prior_only :
         #
         # node_id, split_reference
         node_id            = row['fit_node_id']
         split_reference_id = row['split_reference_id']
         #
         # child_job_table
         child_job_table  = get_child_job_table(
            job_id,
            node_id,
            refit_split,
            split_reference_id,
             root_split_reference_id,
            all_table['split_reference'],
            node_split_set,
            fit_children,
            prior_children,
            node_table,
         )
         #
         # job_table
         row['start_child_job_id'] = len(job_table)
         job_table                += child_job_table
         row['end_child_job_id']   = len(job_table)
      #
      # job_id
      job_id += 1
   #
   # BEGIN_RETURN
   # ...
   assert type(job_table)      == list
   assert type( job_table[0] ) == dict
   assert job_table[0]['fit_node_id'] == start_node_id
   assert job_table[0]['split_reference_id'] == start_split_reference_id
   assert job_table[0]['prior_only'] == False
   for job_id in range(1, len(job_table) ) :
      parent_job_id = job_table[job_id]['parent_job_id']
      assert job_table[parent_job_id]['prior_only'] == False
   return job_table
   # END_RETURN



# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: University of Washington <https://www.washington.edu>
# SPDX-FileContributor: 2021-24 Bradley M. Bell
# ----------------------------------------------------------------------------
import copy
"""
{xrst_begin csv.empty_str}

Create A Table from a CSV File
##############################

Prototype
*********
{xrst_literal ,
   BEGIN_DEF, END_DEF
   BEGIN_RETURN, END_RETURN
}


table_in
********
is a ``list`` of ``dict``.
We use *n* to denote the length of the list which is
the number of lines in the file minus one.
For *i*, an ``int`` ,  between zero and *n* -1,
and each column name *key* , a string,
*table* [ *i* ][ *key* ] is the ``str`` in line *i* +2 and column *key*.

direction
*********
is a ``str`` that is equal to ``to_none`` or ``from_none``.

to_none
=======
If *direction* is to_none, then every empty string is converted to None.
In this case there cannot be any None field values in the input.

from_none
=========
If *direction* is from_none, then every None value is converted to
the empty string.
In this case there cannot be any empty string values in the input.

table_out
*********
Is a copy of *table_in* after doing the conversion.


Example
*******
:ref:`csv.table-name`

{xrst_end csv.empty_str}
"""
# BEGIN_DEF
# at_cascade.csv.empty_str
def empty_str(table_in, direction) :
   assert type(table_in)  == list
   if len(table_in) > 0 :
      assert type(table_in[0]) == dict
   assert direction == 'to_none' or direction == 'from_none'
   # END_DEF
   #
   table_out = list()
   for row in table_in :
      row = copy.copy(row)
      for key in row :
         if direction == 'to_none' :
            if row[key] == '' :
               row[key] = None
         else :
            if row[key] == None :
               row[key] = ''
      table_out.append(row)
   # BEGIN_RETURN
   # ...
   assert type(table_out) == list
   if len(table_out) > 0 :
         assert type(table_out[0]) == dict
   return table_out
   # END_RETURN



# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: University of Washington <https://www.washington.edu>
# SPDX-FileContributor: 2021-24 Bradley M. Bell
# ----------------------------------------------------------------------------
'''
{xrst_begin table_exists}

Check if A Database Table Exists
################################

Prototype
*********
{xrst_literal ,
   # BEGIN_DEF, # END_DEF
   # BEGIN_RETURN, # END_RETURN
}

connection
**********
s a dismod_at connection to the database.


table_name
**********
is the name of the table.

result
******
is either True (table exists) or False (table does not exist).

{xrst_end table_exists}
'''
# -----------------------------------------------------------------------------
import dismod_at
# BEGIN_DEF
# at_cascade.table_exists
def table_exists(
   connection,
   table_name
) :
   assert type(table_name) == str
   # END_DEF
   command = f"SELECT name FROM sqlite_master WHERE name='{table_name}'"
   result  = dismod_at.sql_command(connection, command)
   exists  = len( result ) > 0
   # BEGIN_RETURN
   # ...
   assert type(exists) == bool
   return exists
   # END_RETURN



# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: University of Washington <https://www.washington.edu>
# SPDX-FileContributor: 2021-24 Bradley M. Bell
# ----------------------------------------------------------------------------
'''
{xrst_begin move_table}
{xrst_spell
  dst
  src
}

Move Table
##########

Prototype
*********
{xrst_literal
   # BEGIN_DEF
   # END_DEF
}

Purpose
*******
Change the name of a table in the database.

connection
**********
Is a dismod_at open connection to the database.

src_name
********
is a ``str`` containing the name of the table before moving it.

dst_name
********
is a ``str`` containing the name of the table after moving it.

Log Table
*********
As row is added to the log table with the
ref:`add_log_entry.message` that says that table *src_name* was moved
to table *dst_name*.

1. *log_id*: is the integer length of the log table before the message
2. *message_type* is text ``at_cascade``
3. *table_name* is null
4. *row_id* is null
5. *seconds* is the integer unit time
6. *message* is the text message

{xrst_end   move_table}
'''
import dismod_at
import at_cascade
# ----------------------------------------------------------------------------
# BEGIN_DEF
# at_cascade.move_table
def move_table(
   connection, src_name, dst_name
) :
   assert type(src_name) == str
   assert type(dst_name) == str
   # END_DEF
   #
   command     = 'DROP TABLE IF EXISTS ' + dst_name
   dismod_at.sql_command(connection, command)
   #
   command     = 'ALTER TABLE ' + src_name + ' RENAME COLUMN '
   command    += src_name + '_id TO ' + dst_name + '_id'
   dismod_at.sql_command(connection, command)
   #
   command     = 'ALTER TABLE ' + src_name + ' RENAME TO ' + dst_name
   dismod_at.sql_command(connection, command)
   #
   # log table
   message      = f'move table {src_name} to {dst_name}'
   at_cascade.add_log_entry(connection, message)



# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: University of Washington <https://www.washington.edu>
# SPDX-FileContributor: 2021-25 Bradley M. Bell
# ----------------------------------------------------------------------------
r'''
{xrst_begin get_database_dir}

Get Database Directory Corresponding To a Fit
#############################################

Prototype
*********
{xrst_literal ,
   # BEGIN_DEF, # END_DEF
   # BEGIN_RETURN, # END_RETURN
}

node_table
**********
is the node_table for this cascade as a ``list'' of ``dict``.
It can't be ``None``.

split_reference_table
*********************
is the :ref:`split_reference_table-name` as a ``list`` of ``dict``.
It can't be ``None``.
If the list has length zero,
we say that the table is empty.

node_split_set
**************
If :ref:`split_reference_table-name` is empty,
this argument must be None.
Otherwise it is a ``set`` of ``int`` containing the
:ref:`node_split_table@node_id` values that appear in the
node_spit table.

root_node_id
************
is the node_id for the :ref:`glossary@root_node`.

root_split_reference_id
***********************
If :ref:`split_reference_table-name` is empty,
this argument must be None.
Otherwise it is an ``int`` specifying the
:ref:`split_reference_table@split_reference_id`
that the root node corresponds to.

fit_node_id
***********
This argument is an ``int`` is the node_id for the  :ref:`glossary@fit_node`.

fit_split_reference_id
**********************
If :ref:`split_reference_table-name` is empty,
this argument must be None.
Otherwise it is an ``int`` specifying the
:ref:`split_reference_table@split_reference_id`
that the fit corresponds to.

database_dir
************
The return value is a ``str`` containing the directory,
relative to the :ref:`option_all_table@result_dir`,
where the database corresponding to the fit is (or will be) located.
In other words, the fit database has the following path:

   *result_dir*\ ``/`` *database_dir*\ ``/dismod.db``


{xrst_end get_database_dir}
'''
import dismod_at
# ----------------------------------------------------------------------------
# BEGIN_DEF
# at_cascade.get_database_dir
def get_database_dir(
   node_table                    ,
   split_reference_table         ,
   node_split_set          = None,
   root_node_id            = None,
   root_split_reference_id = None,
   fit_node_id             = None,
   fit_split_reference_id  = None,
) :
   assert type(node_table) == list
   assert type(split_reference_table) == list
   assert node_split_set == None or type(node_split_set) == set
   assert type(root_node_id) == int
   assert root_split_reference_id==None or type(root_split_reference_id)==int
   assert type(fit_node_id) == int
   assert fit_split_reference_id==None or type(fit_split_reference_id)==int
   # END_DEF
   #
   # fit_split_reference_name
   if 0 < len(split_reference_table) :
      row = split_reference_table[fit_split_reference_id]
      fit_split_reference_name = row['split_reference_name']
   #
   # database_dir
   database_dir = ''
   #
   # node_id, split_reference_id
   node_id            = fit_node_id
   split_reference_id = fit_split_reference_id
   #
   while node_id not in [ None, root_node_id ] :
      #
      # split
      split = root_split_reference_id != split_reference_id \
         and node_id in node_split_set
      #
      # next node_id, split_reference_id pair
      if split :
         assert 0 < len(split_reference_table)
         #
         # database_dir
         database_dir      = f'{fit_split_reference_name}/{database_dir}'
         #
         # split_reference_id
         split_reference_id = root_split_reference_id
      else :
         #
         # database_dir
         node_name     = node_table[node_id]['node_name']
         database_dir = f'{node_name}/{database_dir}'
         #
         # node_id
         node_id       = node_table[node_id]['parent']
   if node_id is None :
      fit_node_name  = node_table[fit_node_id]['node_name']
      root_node_name = node_table[root_node_id]['node_name']
      msg  = f'{fit_node_name} is not a descendant of the root node '
      msg += root_node_name
      assert False, msg
   #
   split =  root_split_reference_id != split_reference_id \
      and root_node_id in node_split_set
   if split :
      assert 0 < len(split_reference_table)
      #
      # database_dir
      database_dir = f'{fit_split_reference_name}/{database_dir}'
   #
   root_node_name = node_table[root_node_id]['node_name']
   database_dir = f'{root_node_name}/{database_dir}'
   #
   # database_dir
   database_dir = database_dir[:-1]
   #
   # BEGIN_RETURN
   #  ...
   assert type(database_dir) == str
   return database_dir
   # END_RETURN



# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: University of Washington <https://www.washington.edu>
# SPDX-FileContributor: 2021-24 Bradley M. Bell
# ----------------------------------------------------------------------------
'''
{xrst_begin copy_root_db}

Copy Root Node Database
#######################

Prototype
*********
{xrst_literal
   # BEGIN_COPY_ROOT_DB
   # END_COPY_ROOT_DB
}

root_database
*************
This is the :ref:`glossary@root_database` .
It must exist when this routine is called.

fit_database
************
This is a :ref:`glossary@fit_database` that
can be used to fit the root node.
The directory where the *fit_database* will be located must exist
when this routine is called.
This database is created as follows:

#. Copy the root node database to the fit node database.

#. Drop all of the constant tables from the fit node database; see
   :ref:`module@at_cascade.constant_table_list` .

#. Change the fit node database option table so that is uses the
   root node database for all the constant tables; i.e.,
   set the other_database and other_input_table options to do this.

#. Create an empty log table in the fit node database.


{xrst_end copy_root_db}
'''
import os
import shutil
import at_cascade
import dismod_at
# -----------------------------------------------------------------------------
def create_empty_log_table(connection) :
   #
   cmd  = 'create table if not exists log('
   cmd += ' log_id        integer primary key,'
   cmd += ' message_type  text               ,'
   cmd += ' table_name    text               ,'
   cmd += ' row_id        integer            ,'
   cmd += ' unix_time     integer            ,'
   cmd += ' message       text               )'
   dismod_at.sql_command(connection, cmd)
   #
   # log table
   empty_list = list()
   dismod_at.replace_table(connection, 'log', empty_list)
# -----------------------------------------------------------------------------
# BEGIN_COPY_ROOT_DB
# at_cascade.copy_root_db
def copy_root_db(root_database, fit_database) :
   assert type(root_database) == str
   assert type(fit_database) == str
   # END_COPY_ROOT_DB
   #
   # fit_database
   shutil.copyfile(root_database, fit_database)
   #
   # connection
   connection = dismod_at.create_connection(
      fit_database, new = False, readonly = False
   )
   #
   # fit_database
   for table_name in at_cascade.constant_table_list :
      if table_name == 'rate_eff_cov' :
         command = f'DROP TABLE IF EXISTS {table_name}'
      else :
         command = f'DROP TABLE {table_name}'
      dismod_at.sql_command(connection, command)
   command = 'VACUUM;'
   dismod_at.sql_command(connection, command)
   #
   # other_input_table
   other_input_table = ' '.join(at_cascade.constant_table_list)
   if other_input_table == '' :
      other_input_table = None
   #
   # option_table
   option_table = dismod_at.get_table_dict(connection, tbl_name = 'option')
   for row in option_table :
      assert row['option_name'] != 'other_database'
      assert row['option_name'] != 'other_input_table'
   row = dict()
   row['option_name']  = 'other_database'
   if os.path.isabs( root_database ) :
      row['option_value'] = root_database
   else :
      fit_node_dirname = os.path.dirname( fit_database )
      relative_path    = os.path.relpath(root_database, fit_node_dirname)
      row['option_value'] = str( relative_path )
   option_table.append(row)
   row = dict()
   row['option_name']  = 'other_input_table'
   row['option_value'] = other_input_table
   option_table.append(row)
   #
   # fit_database
   dismod_at.replace_table(
      connection, tbl_name = 'option', table_dict = option_table
   )
   #
   # log table
   create_empty_log_table(connection)
   #
   # connection
   connection.close()



# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: University of Washington <https://www.washington.edu>
# SPDX-FileContributor: 2021-24 Bradley M. Bell
# ----------------------------------------------------------------------------
'''
{xrst_begin com_cov_reference}

Compute Covariate Reference Values
##################################
Compute covariate references by averaging values in the data table
for a specific node and split reference value.

Prototype
*********
{xrst_literal ,
   # BEGIN_DEF, # END_DEF
   # BEGIN_RETURN, # END_RETURN
}

option_all_table
****************
The :ref:`option_all_table@split_covariate_name` and
:ref:`option_all_table@absolute_covariates` rows of this table
(if they exist) are the only rows of this table that are used.

split_reference_table
*********************
is the :ref:`split_reference_table-name` as a ``list``
of ``dict`` .

node_table
**********
This is a list of dict representing the node table in the
:ref:`glossary@root_database` .

covariate_table
***************
This is the covariate table for any fit or the one in the
root node database.

shift_node_id
*************
This is the dismod_at node that the computed
covariate reference values correspond to.

split_reference_id
******************
This is the :ref:`split_reference_table@split_reference_id` that the
computed covariate reference values correspond to.

cov_reference_list
******************
1. The return value is a ``list`` with length equal to the
   length of the covariate table (in the root node database).
2. The :ref:`option_all_table@absolute_covariates` have the same
   reference value as in the :ref:`glossary@root_database` .
3. The splitting covariate has reference value corresponding to
   *split_reference_id* in the split_reference table.
4. The :ref:`glossary@Relative Covariate` reference values are equal to
   the average of the corresponding covariates in the data table.
5. Only rows of the data table that get included when *shift_node_id*
   is the parent node are included in the average.
6. Only rows of the data that are within the max difference for the splitting
   covariate for this *split_reference_id* are included in the average.
7. null values for a covariate are not included in the average.
8. If there are no values to average for a relative covariate, the reference
   in the root_database covariate table is used for that covariate.

{xrst_end com_cov_reference}
'''
import at_cascade
import dismod_at
import math
#
# BEGIN_DEF
# at_cascade.com_cov_reference
def com_cov_reference(
   option_all_table      ,
   split_reference_table ,
   node_table            ,
   covariate_table       ,
   shift_node_id         ,
   split_reference_id    = None,
   data_table            = None,
) :
   assert type(option_all_table) == list
   assert type(split_reference_table) == list
   assert type(node_table) == list
   assert type(covariate_table) == list
   assert type(shift_node_id) == int
   assert type(split_reference_id) == int or split_reference_id == None
   assert type(data_table) == list or data_table == None
   # END_DEF
   #
   # root_database
   root_database      = None
   for row in option_all_table :
      if row['option_name'] == 'root_database' :
         root_database      = row['option_value']
   assert root_database != None
   #
   # check split_reference_id
   if len( split_reference_table ) == 0 :
      assert split_reference_id == None
   else :
      assert type(split_reference_id) == int
   #
   # data_table
   if data_table == None :
      connection = dismod_at.create_connection(
         root_database, new = False, readonly = True
      );
      data_table = dismod_at.get_table_dict(connection, 'data')
      connection.close()
   #
   # cov_info
   cov_info = at_cascade.get_cov_info(
      option_all_table,
      covariate_table,
      split_reference_table
   )
   #
   # rel_covariate_id_set
   rel_covariate_id_set = cov_info['rel_covariate_id_set']
   #
   # split_covariate_id
   split_covariate_id = None
   if len( split_reference_table ) > 0 :
      split_covariate_id = cov_info['split_covariate_id']
   #
   # check max_difference
   for covariate_id in rel_covariate_id_set :
      covariate_row  = covariate_table[covariate_id]
      max_difference = covariate_row['max_difference']
      if not max_difference in [ None, math.inf ] :
         msg  = f'com_cov_reference: covariate_id = {covariate_id}\n'
         msg += 'is a relative covariate and '
         msg += f'max_difference = {max_difference} is not None or infinity'
         assert False, msg
   #
   # n_covariate
   n_covariate = len( covariate_table )
   #
   # covariate_label
   covariate_label = list()
   for covariate_id in range( n_covariate ) :
      covariate_label.append( f'x_{covariate_id}' )
   #
   # is_descendant
   is_descendant = set()
   for (node_id, row) in enumerate(node_table) :
      this_is_descendant = node_id == shift_node_id
      ancestor_node_id   = row['parent']
      while not ancestor_node_id is None :
         if ancestor_node_id == shift_node_id :
            this_is_descendant = True
         ancestor_row     = node_table[ancestor_node_id]
         ancestor_node_id = ancestor_row['parent']
      if this_is_descendant :
         is_descendant.add( node_id )
   #
   # split_reference_value
   if len( split_reference_table ) > 0 :
      row  = split_reference_table[split_reference_id]
      split_reference_value = row['split_reference_value']
   #
   # data_subset_list
   data_subset_list = list()
   for (data_id, data_row) in enumerate(data_table) :
      #
      # node_id
      node_id = data_row['node_id']
      if node_id in is_descendant :
         #
         # in_bnd
         in_bnd = True
         for covariate_id in range( n_covariate ) :
            covariate_row   = covariate_table[covariate_id]
            reference       = covariate_row['reference']
            if covariate_id == split_covariate_id :
               reference = split_reference_value
            max_difference  = covariate_row['max_difference']
            if max_difference is None :
               max_difference = math.inf
            label           = covariate_label[covariate_id]
            covariate_value = data_row[label]
            #
            skip = covariate_value is None
            skip = skip or max_difference == math.inf
            if not skip :
               abs_diff = abs( covariate_value - reference )
               in_bnd   = in_bnd and abs_diff <= max_difference
         #
         # data_subset_list
         if in_bnd :
            data_subset_list.append( data_id )
   #
   # cov_reference_list
   cov_reference_list = list()
   for covariate_id in range( n_covariate) :
      #
      # reference
      reference = covariate_table[covariate_id]['reference']
      if covariate_id == split_covariate_id :
         reference = split_reference_value
      #
      if covariate_id in rel_covariate_id_set :
         #
         # covariate_list
         covariate_row  = covariate_table[covariate_id]
         covariate_list = list()
         for data_id in data_subset_list :
            data_row  = data_table[data_id]
            cov_value = data_row[ covariate_label[covariate_id] ]
            if not cov_value is None :
               covariate_list.append(cov_value)
         #
         # reference
         if len( covariate_list ) > 0 :
            reference = sum(covariate_list) / len(covariate_list)
      #
      # cov_reference_list
      cov_reference_list.append(reference)
   # -------------------------------------------------------------------------
   # BEGIN_RETURN
   # ...
   assert type(cov_reference_list) == list
   return cov_reference_list
   # END_RETURN



# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: University of Washington <https://www.washington.edu>
# SPDX-FileContributor: 2021-25 Bradley M. Bell
# ----------------------------------------------------------------------------
r'''
{xrst_begin fit_one_job}
{xrst_spell
  obj
  var
}

Run One Job
###########

Prototype
*********
{xrst_literal
   # BEGIN_DEF
   # END_DEF
}

Default Value
*************
The only argument that can be None is *trace_file_obj*.

job_table
*********
This is a :ref:`create_job_table@job_table` containing the jobs
necessary to fit the :ref:`glossary@fit_goal_set`.

run_job_id
**********
This is the :ref:`create_job_table@job_table@job_id`
for the job that is run.
If *run_job_id* is zero, this job has no ancestors.
This is a special case for :ref:`fit_one_job@fit_database@fit_var`

all_node_database
*****************
is a python string specifying the location of the
:ref:`all_node_db-name`
relative to the current working directory.

node_table
**********
is a ``list`` of ``dict`` containing the node table for this cascade.

fit_integrand
*************
is a ``set`` of integrand_id values that occur in the data table; see
:ref:`get_fit_integrand-name`.

fit_type
********
is a ``str`` equal to 'both' or 'fixed' and specifies the type
of fit that dismod_at will do.

first_fit
*********
If *first_fit* is True, this is assumed to be an
:ref:`glossary@input_node_database`.
Otherwise, it is assumed that this
routine has previously been called with *first_fit* equal to True.

trace_file_obj
**************
If this argument is not None, it is a ``io.TextIOBase`` object
corresponding to a file that is opened for writing the tracing output
for this job.

fit_database
************
The :ref:`glossary@fit_database` for this fit is
*fit_node_dir*\ ``/dismod.db`` where *fit_node_dir*
is the :ref:`get_database_dir@database_dir` returned by
get_database_dir for the fit node and split_reference_id corresponding
to *run_job_id*.

Upon Input
==========
On input, *fit_database* is an :ref:`glossary@input_node_database`.

fit_var
=======
Upon return, the fit_var table correspond to the posterior
mean for the model variables for the fit_node.
If :ref:`fit_one_job@run_job_id` is not zero
and there is no data corresponding to this fit,
the fit is not done because an ancestor job can be used to predict for this job.
In this case a no data abort message will appear in the
:ref:`fit_one_job@fit_database@log` table.

sample
======
Upon return,
the sample table contains the corresponding samples from the posterior
distribution for the model variables for the fit_node.

log
===
The log table is initialized as empty when ``fit_one_job`` starts.
Upon return or abort due to an exception,
the log table contains a summary of the operations preformed by dismod.
In addition,
the following entries will be added to the log table
if the corresponding event occurs:

.. csv-table::
   :header-rows:1

   message_type, message,        event
   at_cascade,   no data: abort, abort fit because all the data is held out
   at_cascade,   fit: OK,        the maximum likelihood problem was solved
   at_cascade,   sample: OK,     the posterior samples were computed
   at_cascade,   children: OK,   the child databases with priors were created

Note that the events depend on each other in the following way:

#. If children: OK is present, then sample: OK is present.
#. If sample: OK is present, then fit: OK is present.
#. If fit: OK is present, then no data: abort is **not** present.


Exception
*********
If there is no data from this fit, this routine will raise an exception
with a message that starts with: ``no data: abort`` ; i.e., the same
as the message it puts in the log.


{xrst_end fit_one_job}
'''
# ----------------------------------------------------------------------------
import io
import os
import time
import inspect
import dismod_at
import at_cascade
# -----------------------------------------------------------------------------
def system_command(command, file_stdout) :
   if file_stdout is None :
      dismod_at.system_command_prc(
         command,
         print_command = True,
         return_stdout = False,
         return_stderr = False,
         file_stdout   = None,
         file_stderr   = None,
         write_command = False,
      )
   else :
      dismod_at.system_command_prc(
         command,
         print_command = False,
         return_stdout = False,
         return_stderr = False,
         file_stdout   = file_stdout,
         file_stderr   = None,
         write_command = True,
      )
# ----------------------------------------------------------------------------
# BEGIN_DEF
# at_cascade.fit_one_job
def fit_one_job(
   job_table               ,
   run_job_id              ,
   all_node_database       ,
   node_table              ,
   fit_integrand           ,
   fit_type                ,
   first_fit               ,
   trace_file_obj   = None ,
) :
   assert type(job_table) == list
   assert type(run_job_id) == int
   assert type(all_node_database) == str
   assert type(node_table) == list
   assert type(fit_integrand) == set
   assert fit_type in [ 'both', 'fixed' ]
   assert type(first_fit) == bool
   if trace_file_obj is not None :
      assert isinstance(trace_file_obj, io.TextIOBase)
   # END_DEF
   #
   # trace_line_number
   # You can use this routine to help track down a crash during fit_one_job.
   def trace_line_number(line_number ) :
      msg = f'fit_one_job.py: line {line_number}'
      if trace_file_obj == None :
         print( msg )
      else :
         trace_file_obj.write( f'{msg}\n' )
         trace_file_obj.flush()
   # trace_line_number( inspect.currentframe().f_lineno )
   #
   # file_stdout
   file_stdout = trace_file_obj
   #
   # fit_node_id
   fit_node_id = job_table[run_job_id]['fit_node_id']
   #
   # fit_split_reference_id
   fit_split_reference_id = job_table[run_job_id]['split_reference_id']
   #
   # start_child_job_id
   start_child_job_id = job_table[run_job_id]['start_child_job_id']
   #
   # end_child_job_id
   end_child_job_id = job_table[run_job_id]['end_child_job_id']
   #
   # all_table
   connection  = dismod_at.create_connection(
      all_node_database, new = False, readonly = True
   )
   all_table = dict()
   for tbl_name in [
      'option_all',
      'split_reference',
      'node_split',
      'mulcov_freeze',
   ] :
      all_table[tbl_name] = dismod_at.get_table_dict(connection, tbl_name)
   connection.close()
   #
   # double_max_fit
   double_max_fit = False
   for row in all_table['mulcov_freeze'] :
      if fit_node_id == row['fit_node_id'] :
         if fit_split_reference_id == row['split_reference_id'] :
            double_max_fit = True
   #
   # option_all_dict
   option_all_dict = dict()
   for row in all_table['option_all'] :
      option_all_dict[ row['option_name']  ] = row['option_value']
   #
   # sample_method
   if 'sample_method' in option_all_dict :
      sample_method = option_all_dict['sample_method']
      if sample_method not in [
         'asymptotic', 'censor_asymptotic', 'simulate' ] :
         msg  = 'option_all table: sample_method is not '
         msg += 'asymptotic , censor_asymptotic, or simulate'
         assert False, msg
   else :
      sample_method = 'asymptotic'
   #
   # refit_split
   if 'refit_split' in option_all_dict :
      refit_split = option_all_dict['refit_split']
      assert refit_split in [ 'true', 'false' ]
      refit_split = refit_split == 'true'
   else :
      refit_split = False
   #
   # result_dir
   result_dir = option_all_dict['result_dir']
   #
   # root_node_id
   name         = option_all_dict['root_node_name']
   root_node_id = at_cascade.table_name2id(node_table, 'node', name)
   #
   # root_split_reference_id
   if 'root_split_reference_name' not in option_all_dict :
      root_split_reference_id = None
      assert refit_split == False
   else :
      name = option_all_dict['root_split_reference_name']
      root_split_reference_id   = at_cascade.table_name2id(
         all_table['split_reference'], 'split_reference', name
      )
   #
   # max_fit_parent
   if 'max_fit_parent' not in option_all_dict :
      max_fit_parent = None
   else :
      max_fit_parent = option_all_dict['max_fit_parent']
      if 'max_fit' not in option_all_dict :
         msg  = 'max_fit_parent appears in option_all table '
         msg += 'but max_fit does not.'
         assert False, msg
   #
   # balance_fit
   if 'balance_fit' not in option_all_dict :
      balance_fit = None
   else :
      balance_fit = option_all_dict['balance_fit']
      balance_fit = balance_fit.split()
      if 'max_fit' not in option_all_dict :
         msg  = 'balance_fit appears in option_all table '
         msg += 'but max_fit does not.'
         assert False, msg
      if len(balance_fit) != 3 :
         msg = 'option_all table: balance_fit is not a space separated '
         msg += 'list with three elements'
         assert False, msg
   #
   # node_split_set
   node_split_set = set()
   for row in all_table['node_split'] :
      node_split_set.add( row['node_id'] )
   #
   # fit_database
   database_dir = at_cascade.get_database_dir(
      node_table              = node_table,
      split_reference_table   = all_table['split_reference'],
      node_split_set          = node_split_set,
      root_node_id            = root_node_id,
      root_split_reference_id = root_split_reference_id,
      fit_node_id             = fit_node_id ,
      fit_split_reference_id  = fit_split_reference_id,
   )
   fit_database      = f'{result_dir}/{database_dir}/dismod.db'
   #
   # check fit_database
   parent_node_name = at_cascade.get_parent_node(fit_database)
   assert parent_node_name == node_table[fit_node_id]['node_name']
   #
   # integrand_table
   root_database      = option_all_dict['root_database']
   fit_or_root        = at_cascade.fit_or_root_class(
      fit_database, root_database
   )
   integrand_table = fit_or_root.get_table('integrand')
   fit_or_root.close()
   #
   # dismod_at_version
   dismod_at_version = dismod_at.system_command_prc(
      command       = [ 'dismod_at', '--version' ] ,
      print_command = False,
      return_stdout = True,
      return_stderr = False,
   )
   dismod_at_version = dismod_at_version.strip()
   #
   # at_cascade_version
   at_cascade_version = 'at_cascade-' + at_cascade.version
   #
   # fit_database: log table
   connection = dismod_at.create_connection(
      fit_database, new = False, readonly = False
   )
   command = 'DROP TABLE IF EXISTS log'
   dismod_at.sql_command(connection, command)
   at_cascade.add_log_entry(connection, dismod_at_version)
   at_cascade.add_log_entry(connection, at_cascade_version)
   connection.close()
   #
   # init
   command = [ 'dismod_at', fit_database, 'init' ]
   system_command(command, file_stdout)
   #
   # max_fit
   if 'max_fit' in option_all_dict :
      max_fit = option_all_dict['max_fit']
      if double_max_fit :
         max_fit = str( 2 * int(max_fit) )
      for integrand_id in fit_integrand :
         integrand_name = integrand_table[integrand_id]['integrand_name']
         command = [
            'dismod_at', fit_database,
            'hold_out', integrand_name, max_fit
         ]
         if max_fit_parent is not None :
            command += [ max_fit_parent ]
         if balance_fit is not None :
            command += balance_fit
         system_command(command, file_stdout)
   #
   # max_abs_effect
   if 'max_abs_effect' in option_all_dict:
      max_abs_effect = option_all_dict['max_abs_effect']
      command =[
         'dismod_at', fit_database, 'bnd_mulcov', max_abs_effect
      ]
      system_command(command, file_stdout)
   #
   # perturb_optimization
   perturb_optimization = dict()
   for key in [ 'start', 'scale' ] :
      long_key = f'perturb_optimization_{key}'
      if long_key in option_all_dict :
         sigma = option_all_dict[long_key]
         if float(sigma) < 0.0 :
            msg = f'fit_one_job: perturb_optimization_{key} = '
            msg += sigma
            msg += ' is less than zero'
            assert False, msg
         if float(sigma) > 0.0 :
            perturb_optimization[key] = sigma
   #
   # fit_database: scale_var and start_var tables
   for key in perturb_optimization :
      sigma    = perturb_optimization[key]
      tbl_name = f'{key}_var'
      dismod_at.perturb_command( fit_database, tbl_name, sigma )
   #
   # fit_node_datase.log_table
   # if fit has no data, abort with 'fit: error: no data abort' in log_table
   # ( unless this fit has no ancestors; i.e., run_job_id == 0 ).
   data_include_table = at_cascade.data_include(
      fit_database, root_database
   )
   if len( data_include_table )  == 0 and run_job_id > 0:
      msg        = 'no data: abort'
      connection = dismod_at.create_connection(
         fit_database, new = False, readonly = False
      )
      at_cascade.add_log_entry(connection, msg)
      connection.close()
      #
      job_name = job_table[run_job_id]['job_name']
      msg      = f'no data: abort {job_name}'
      raise Exception(msg)
   #
   # fit
   command = [ 'dismod_at', fit_database, 'fit', fit_type ]
   system_command(command, file_stdout)
   #
   # fit_database.log_table
   connection = dismod_at.create_connection(
      fit_database, new = False, readonly = False
   )
   msg      = 'fit: OK'
   at_cascade.add_log_entry(connection, msg)
   connection.close()
   #
   # number_simulate
   if 'number_sample' not in option_all_dict :
      number_simulate = '20'
   else :
      number_simulate = option_all_dict['number_sample']
   #
   # sample
   if sample_method == 'simulate' :
      if int( number_simulate ) > 20 :
         msg  = 'option_all table: number_sample > 20 and '
         msg += 'sample_method is simulate.'
         assert False, msg
      command = [
         'dismod_at', fit_database, 'set', 'truth_var', 'fit_var'
      ]
      system_command(command, file_stdout)
      command = [
         'dismod_at', fit_database, 'simulate', number_simulate
      ]
      system_command(command, file_stdout)
   command = [
      'dismod_at',
      fit_database,
      'sample',
      sample_method,
      fit_type,
      number_simulate
   ]
   system_command(command, file_stdout)
   #
   # fit_database.log_table
   connection = dismod_at.create_connection(
      fit_database, new = False, readonly = False
   )
   msg      = 'sample: OK'
   at_cascade.add_log_entry(connection, msg)
   connection.close()
   #
   # avgint_parent_grid
   at_cascade.avgint_parent_grid(
      all_node_database = all_node_database ,
      fit_database      = fit_database ,
      job_table         = job_table         ,
      fit_job_id        = run_job_id        ,
   )
   #
   # connection
   connection = dismod_at.create_connection(
      fit_database, new = False, readonly = False
   )
   #
   # c_shift_predict_fit_var
   command = [ 'dismod_at', fit_database, 'predict', 'fit_var' ]
   system_command(command, file_stdout)
   at_cascade.move_table(connection, 'predict', 'c_shift_predict_fit_var')
   #
   # c_shift_predict_sample
   command = [ 'dismod_at', fit_database, 'predict', 'sample' ]
   system_command(command, file_stdout)
   at_cascade.move_table(connection, 'predict', 'c_shift_predict_sample')
   #
   # c_shift_avgint
   # is the table created by avgint_parent_grid
   at_cascade.move_table(connection, 'avgint', 'c_shift_avgint')
   #
   # connection
   connection.close()
   #
   # shift_databases
   shift_databases = dict()
   for job_id in range(start_child_job_id, end_child_job_id) :
      #
      # shift_node_id
      shift_node_id = job_table[job_id]['fit_node_id']
      #
      # shift_split_reference_id
      shift_split_reference_id = job_table[job_id]['split_reference_id']
      #
      # shift_database_dir
      database_dir = at_cascade.get_database_dir(
         node_table              = node_table,
         split_reference_table   = all_table['split_reference'],
         node_split_set          = node_split_set,
         root_node_id            = root_node_id,
         root_split_reference_id = root_split_reference_id,
         fit_node_id             = shift_node_id ,
         fit_split_reference_id  = shift_split_reference_id,
      )
      shift_database_dir = f'{result_dir}/{database_dir}'
      if not os.path.exists(shift_database_dir) :
         os.makedirs(shift_database_dir)
      #
      # shift_node_database
      shift_node_database = f'{shift_database_dir}/dismod.db'
      #
      # skip_refit
      if refit_split :
         skip_refit = False
      else :
         skip_refit = fit_split_reference_id != shift_split_reference_id
      #
      # shift_name
      dir_list = shift_database_dir.split('/')
      if skip_refit :
         shift_name = dir_list[-2] + '/' + dir_list[-1]
      else :
         shift_name = dir_list[-1]
      #
      # shift_databases
      shift_databases[shift_name] = shift_node_database
   #
   # create shifted databases
   at_cascade.create_shift_db(
      all_node_database = all_node_database,
      fit_database      = fit_database,
      shift_databases   = shift_databases,
      no_ode_fit        = False,
      job_table         = job_table,
   )
   #
   # empty_avgint_table
   connection = dismod_at.create_connection(
      fit_database, new = False, readonly = False
   )
   at_cascade.empty_avgint_table(connection)
   connection.close()
   #
   #
   # fit_database.log_table
   connection = dismod_at.create_connection(
      fit_database, new = False, readonly = False
   )
   msg      = 'children: OK'
   at_cascade.add_log_entry(connection, msg)
   connection.close()
   #
   # trace_line_number( inspect.currentframe().f_lineno )



# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: University of Washington <https://www.washington.edu>
# SPDX-FileContributor: 2021-24 Bradley M. Bell
# ----------------------------------------------------------------------------
# Set this to False when debugging an exception during fit_one_job routine
catch_exceptions_and_continue = True
# ----------------------------------------------------------------------------
'''
{xrst_begin fit_one_process}
{xrst_spell
  cpus
  dtype
  inuse
  numpy
}

Fit Using One Process
#####################
Keep fitting jobs for as long as there are jobs ready to be fit.

Prototype
*********
{xrst_literal
   # BEGIN_DEF
   # END_DEF
}

job_table
*********
This is a :ref:`create_job_table@job_table` containing all the jobs
necessary to fit the :ref:`glossary@fit_goal_set` .

this_job_id
***********
if skip_this_job is false,
the fit for this ref:`create_job_table@job_table@job_id` is ready to be fit
and will be fit first.
Otherwise, *this_job_id* is ignored.

all_node_database
*****************
:ref:`fit_one_job@all_node_database`

node_table
**********
:ref:`fit_one_job@node_table`

fit_integrand
*************
:ref:`fit_one_job@fit_integrand`

skip_this_job
*************
see :ref:`fit_one_process@this_job_id` above.

max_number_cpu
**************
This is the maximum number of cpus (processes) to use.
This must be greater than zero.
If it is one, the jobs are run sequentially; i.e., not in parallel.

fit_type_list
*************
This is a list with one or two elements
and its possible elements are ``both`` and ``fixed``.
For each job, the first type of fit is attempted.
If it fails, and there is a second type of fit, it is attempted.
If it also fails, the corresponding job fails.

job_status_name
***************
is the name corresponding to each possible job status integer values.
If *i* is an integer job status value, *job_status_name* [ *i* ] is the
corresponding name.

.. csv-table::
   :header-rows: 1

   Name,    Meaning
   'skip' , This is a prior only job and completed by the parent fit
   'wait',  job is waiting for it's parent job to finish
   'ready', job is ready to run
   'run',   job is running
   'done',  job finished running
   'error', job had an exception
   'abort', job is a descendant of a job that had an exception

shared_job_status_name
**********************
This is the name of the shared job status memory.
The corresponding multiprocessing shared memory is
a numpy array with ``dtype`` equal to ``int`` and
with length equal to the length of *job_table* .
The value *shared_job_status* [ *job_table_index* ] is the
integer status code for the corresponding job; see *job_status_name* above.

number_cpu_inuse
****************
This is the name of the number of cpus in use memory.
The corresponding multiprocessing shared memory is
a numpy array with ``dtype`` equal to ``int`` and
with length equal to one.
The value *number_cpu_inuse* [0] is the number of cpus (processes)
currently fitting this cascade.

shared_lock
***********
is a shared memory lock, used by all the fit processes,
that must be acquired to read or write shared memory; i.e.,
*shared_job_status* or *shared_number_cpu_in_use* .

shared_event
************
is multiprocessing event,  used by all the fit processes,
that is used to signal that the shared memory has changed.

{xrst_end fit_one_process}
'''
# ----------------------------------------------------------------------------
import sys
import datetime
import multiprocessing
from multiprocessing import shared_memory
import numpy
import at_cascade
import dismod_at
# ----------------------------------------------------------------------------
# acquire lock
def acquire_lock(shared_lock) :
   seconds = 10
   ok = shared_lock.acquire(
      block   = True,
      timeout = seconds
   )
   if not ok :
      msg = f'pre_one_process: did not obtain lock in {seconds} seconds'
      sys.exit(msg)
# ----------------------------------------------------------------------------
def get_result_database_dir(
   all_node_database, node_table, fit_node_id, fit_split_reference_id
) :
   #
   # option_all, node_split_table, split_reference_table
   connection       = dismod_at.create_connection(
      all_node_database, new = False, readonly = True
   )
   option_all_table = dismod_at.get_table_dict(connection, 'option_all')
   node_split_table = dismod_at.get_table_dict(connection, 'node_split')
   split_reference_table = \
      dismod_at.get_table_dict(connection, 'split_reference')
   connection.close()
   #
   # result_dir, root_node_name
   result_dir              = None
   root_node_id            = None
   root_split_reference_id = None
   for row in option_all_table :
      if row['option_name'] == 'result_dir' :
         result_dir = row['option_value']
      if row['option_name'] == 'root_node_name' :
         root_node_name = row['option_value']
         root_node_id   = \
            at_cascade.table_name2id(node_table, 'node', root_node_name)
      if row['option_name'] == 'root_split_reference_name' :
         root_split_reference_name = row['option_value']
         root_split_reference_id = at_cascade.table_name2id(
            split_reference_table,
            'split_reference',
            root_split_reference_name
         )
   assert result_dir is not None
   assert root_node_id is not None
   #
   # node_split_set
   node_split_set = set()
   for row in node_split_table :
      node_split_set.add( row['node_id'] )
   #
   database_dir = at_cascade.get_database_dir(
      node_table              = node_table,
      split_reference_table   = split_reference_table,
      node_split_set          = node_split_set,
      root_node_id            = root_node_id,
      root_split_reference_id = root_split_reference_id,
      fit_node_id             = fit_node_id,
      fit_split_reference_id  = fit_split_reference_id,
   )
   return f'{result_dir}/{database_dir}'
# )
# ----------------------------------------------------------------------------
def try_one_job(
   job_table,
   this_job_id,
   all_node_database,
   node_table,
   fit_integrand,
   skip_this_job,
   max_number_cpu,
   master_process,
   fit_type_list,
   shared_lock,
   shared_event,
   shared_job_status,
   job_status_name,
)  :
   assert type(job_table) == list
   assert type(this_job_id) == int
   assert type(all_node_database) == str
   assert type(node_table) == list
   assert type(fit_integrand) == set
   assert type(skip_this_job) == bool
   assert type(max_number_cpu) == int
   assert type(master_process) == bool
   assert type(fit_type_list) == list
   #
   # job_status_name
   job_status_skip  = job_status_name.index( 'skip' )
   job_status_wait  = job_status_name.index( 'wait' )
   job_status_ready = job_status_name.index( 'ready' )
   job_status_run   = job_status_name.index( 'run' )
   job_status_done  = job_status_name.index( 'done' )
   job_status_error = job_status_name.index( 'error' )
   job_status_abort = job_status_name.index( 'abort' )
   #
   # database_dir
   row = job_table[this_job_id]
   fit_node_id            = row['fit_node_id']
   fit_split_reference_id = row['split_reference_id']
   result_database_dir = get_result_database_dir(
      all_node_database,
      node_table,
      fit_node_id,
      fit_split_reference_id
   )
   #
   # job_name
   job_name = job_table[this_job_id]['job_name']
   #
   # prior_only
   assert not job_table[this_job_id]['prior_only']
   #
   # trace_file_obj
   trace_file_obj = None
   if max_number_cpu > 1 :
      trace_file_name = f'{result_database_dir}/trace.out'
      trace_file_obj  = open(trace_file_name, 'w')
   #
   # job_done, fit_type_index, fit_type, have_data
   job_done       = False
   have_data      = True
   fit_type_index = 0
   while have_data and (not job_done) and (fit_type_index< len(fit_type_list)) :
      fit_type        = fit_type_list[fit_type_index]
      fit_type_index += 1
      #
      # print message at start of this fit
      now             = datetime.datetime.now()
      current_time    = now.strftime("%H:%M:%S")
      print( f'Begin: {current_time}: fit {fit_type:<5} {job_name}' )
      #
      # fit_one_job
      # the lock should not be acquired during this operation
      if not catch_exceptions_and_continue :
         at_cascade.fit_one_job(
            job_table         = job_table,
            run_job_id        = this_job_id ,
            all_node_database = all_node_database,
            node_table        = node_table,
            fit_integrand     = fit_integrand,
            fit_type          = fit_type,
            first_fit         = fit_type_index == 1,
            trace_file_obj    = trace_file_obj,
         )
         #
         # job_done
         job_done = True
      else :
         try :
            at_cascade.fit_one_job(
               job_table         = job_table,
               run_job_id        = this_job_id ,
               all_node_database = all_node_database,
               node_table        = node_table,
               fit_integrand     = fit_integrand,
               fit_type          = fit_type,
               first_fit         = fit_type_index == 1,
               trace_file_obj    = trace_file_obj,
            )
            #
            # job_done
            job_done = True
         except Exception as e:
            job_done = False
            msg      = str(e)
            if msg.startswith( 'no data: abort' ) :
               have_data = False
            print( f'fit {fit_type} {job_name} message: ' + msg )
   #
   # trace_file_obj
   if trace_file_obj != None :
      trace_file_obj.close()
   #
   if job_done :
      #
      # shared_lock
      acquire_lock(shared_lock)
      #
      # shared_job_status
      assert shared_job_status[this_job_id] == job_status_run
      shared_job_status[this_job_id] = job_status_done
      #
      # shared_job_status[child_job_id]
      start_child_job_id    = job_table[this_job_id ]['start_child_job_id']
      end_child_job_id      = job_table[this_job_id ]['end_child_job_id']
      child_range = range(start_child_job_id, end_child_job_id)
      for child_job_id in child_range :
         if shared_job_status[child_job_id] == job_status_wait :
            assert not job_table[child_job_id]['prior_only']
            shared_job_status[child_job_id] = job_status_ready
         else :
            assert job_table[child_job_id]['prior_only']
            assert shared_job_status[child_job_id] == job_status_skip
      #
      # release
      # shared memory has changed
      shared_event.set()
      shared_lock.release()
      #
   else :
      # if job not ok
      #
      # descendant_set
      descendant_set = { this_job_id }
      n_job          = len(job_table)
      for job_id in range(this_job_id + 1, n_job) :
         parent_job_id = job_table[job_id]['parent_job_id']
         if parent_job_id in descendant_set :
            descendant_set.add( job_id )
      descendant_set.remove( this_job_id )
      #
      # shared_lock
      acquire_lock(shared_lock)
      #
      # shared_job_status[this_job_id]
      if shared_job_status[this_job_id] != job_status_run :
         msg  = 'try_one_job: except: shared_job_status[this_job_id] = '
         msg += job_status_name[ shared_job_status[this_job_id] ]
         print(msg)
      shared_job_status[this_job_id] = job_status_error
      #
      # shared_job_status[descendant_set]
      for job_id in descendant_set :
         if shared_job_status[job_id] != job_status_skip :
            if shared_job_status[job_id] != job_status_wait :
               msg  = 'try_one_job: except: shared_job_status[job_id] = '
               msg += job_status_name[ shared_job_status[job_id] ]
               print(msg)
            shared_job_status[job_id] = job_status_abort
      #
      # release
      # shared memory has changed
      shared_event.set()
      shared_lock.release()
      #
      # ok
      job_done = False
   #
   if max_number_cpu > 1 :
      #
      # print message at end
      job_name     = job_table[this_job_id]['job_name']
      now          = datetime.datetime.now()
      current_time = now.strftime("%H:%M:%S")
      if job_done :
         print( f'End:   {current_time}: fit {fit_type:<5} {job_name}' )
      else :
         print( f'Error: {current_time}: fit {fit_type:<5} {job_name}' )
      #
      # status_count
      acquire_lock(shared_lock)
      n_status      = len( job_status_name )
      status_count  = dict()
      for job_status_i in range(n_status) :
         name               = job_status_name[job_status_i]
         status_count[name] =  int( sum( shared_job_status == job_status_i ) )
      shared_lock.release()
      #
      print( f'       {status_count}' )
      #
   return
# ----------------------------------------------------------------------------
# BEGIN_DEF
# at_cascade.fit_one_process
def fit_one_process(
   job_table,
   this_job_id,
   all_node_database,
   node_table,
   fit_integrand,
   skip_this_job,
   max_number_cpu,
   master_process,
   fit_type_list,
   job_status_name,
   shared_job_status_name,
   shared_number_cpu_inuse_name,
   shared_lock,
   shared_event,
) :
   assert type(job_table)            == list
   assert type(this_job_id)          == int
   assert type(all_node_database)    == str
   assert type(node_table)           == list
   assert type(fit_integrand)        == set
   assert type(skip_this_job)        == bool
   assert type(max_number_cpu)       == int
   assert type(master_process)       == bool
   assert type(fit_type_list)        == list
   assert type(job_status_name)      == list
   assert type( job_status_name[0] ) == str
   assert type(shared_job_status_name)       == str
   assert type(shared_number_cpu_inuse_name) == str
   assert type(shared_lock)          == multiprocessing.synchronize.Lock
   assert type(shared_event)         == multiprocessing.synchronize.Event
   # END_DEF
   # ----------------------------------------------------------------------
   job_status_skip  = job_status_name.index( 'skip' )
   job_status_wait  = job_status_name.index( 'wait' )
   job_status_ready = job_status_name.index( 'ready' )
   job_status_run   = job_status_name.index( 'run' )
   job_status_done  = job_status_name.index( 'done' )
   job_status_error = job_status_name.index( 'error' )
   job_status_abort = job_status_name.index( 'abort' )
   # ----------------------------------------------------------------------
   #
   # shm_job_status, shared_job_status
   tmp    = numpy.empty(len(job_table), dtype = int )
   mapped = at_cascade.map_shared( shared_job_status_name )
   shm_job_status = multiprocessing.shared_memory.SharedMemory(
      create = False, size = tmp.nbytes, name = mapped
   )
   shared_job_status = numpy.ndarray(
      tmp.shape, dtype = tmp.dtype, buffer = shm_job_status.buf
   )
   #
   # shm_number_cpu_inuse, shared_number_cpu_inuse
   tmp    = numpy.empty(1, dtype = int )
   mapped = at_cascade.map_shared( shared_number_cpu_inuse_name )
   shm_number_cpu_inuse = multiprocessing.shared_memory.SharedMemory(
      create = False, size = tmp.nbytes, name = mapped
   )
   shared_number_cpu_inuse = numpy.ndarray(
      tmp.shape, dtype = tmp.dtype, buffer = shm_number_cpu_inuse.buf
   )
   #
   # job_table_index
   job_table_index = numpy.array( range(len(job_table)), dtype = int )
   #
   if not skip_this_job :
      #
      # try_one_job
      # assumes lock not acquired during this operation
      try_one_job(
         job_table,
         this_job_id,
         all_node_database,
         node_table,
         fit_integrand,
         skip_this_job,
         max_number_cpu,
         master_process,
         fit_type_list,
         shared_lock,
         shared_event,
         shared_job_status,
         job_status_name,
      )
   #
   while True :
      # shared_lock
      acquire_lock(shared_lock)
      #
      # job_id_ready
      job_id_ready = job_table_index[ shared_job_status == job_status_ready ]
      #
      # job_id_run
      job_id_run  = job_table_index[ shared_job_status == job_status_run ]
      #
      # n_job_ready
      n_job_ready = job_id_ready.size
      #
      if n_job_ready == 0 :
         if job_id_run.size == 0 :
            #
            # no jobs running or ready
            if master_process :
               if shared_number_cpu_inuse[0] == 1:
                  # We are done, return to fit_parallel which wuill use
                  # the shared memory for error checking and then free it.
                  #
                  # should not need this release
                  shared_lock.release()
                  #
                  shm_job_status.close()
                  shm_number_cpu_inuse.close()
                  return
               else :
                  #
                  # wait for another process to change shared memory,
                  # then go back to the while True point above
                  shared_event.clear()
                  shared_lock.release()
                  seconds = 10.0
                  shared_event.wait(timeout = seconds)
            else :
               # return this processor
               shared_number_cpu_inuse[0] -= 1
               #
               # release
               # shared memory has changed
               shared_event.set()
               shared_lock.release()
               #
               shm_job_status.close()
               shm_number_cpu_inuse.close()
               return
         else :
            #
            # jobs are running but none are ready
            if master_process :
               #
               # wait for another process to change shared memory,
               # then go back to the while True point above
               shared_event.clear()
               shared_lock.release()
               seconds = 10.0
               shared_event.wait(timeout = seconds)
            else :
               # return this processor
               shared_number_cpu_inuse[0] -= 1
               #
               # release
               shared_lock.release()
               #
               shm_job_status.close()
               shm_number_cpu_inuse.close()
               return
      else :
         #
         # n_cpu_spawn
         n_cpu_available  = max_number_cpu - shared_number_cpu_inuse[0]
         n_cpu_spawn      = min(n_cpu_available, n_job_ready - 1)
         #
         # shared_number_cpu_inuse
         shared_number_cpu_inuse[0] += n_cpu_spawn
         #
         # shared_job_status
         for i in range( n_cpu_spawn + 1 ) :
            job_id = job_id_ready[i]
            #
            assert shared_job_status[job_id] == job_status_ready
            shared_job_status[job_id] = job_status_run
         #
         # release
         # shared memory has changed
         shared_event.set()
         shared_lock.release()
         #
         # skip_child_job
         skip_child_job = False
         #
         # is_child_master_process
         is_child_master_process = False
         #
         # spawn the new processes
         for i in range(n_cpu_spawn) :
            #
            # job_id
            job_id = int( job_id_ready[i] )
            #
            # p
            args = (
               job_table,
               job_id,
               all_node_database,
               node_table,
               fit_integrand,
               skip_child_job,
               max_number_cpu,
               is_child_master_process,
               fit_type_list,
               job_status_name,
               shared_job_status_name,
               shared_number_cpu_inuse_name,
               shared_lock,
               shared_event,
            )
            target = fit_one_process
            p = multiprocessing.Process(target = target, args = args)
            #
            p.daemon = False
            p.start()
         #
         # job_id
         job_id = int( job_id_ready[n_cpu_spawn] )
         #
         # try_one_job
         # assumes lock is not acquired during this operation
         try_one_job(
            job_table,
            job_id,
            all_node_database,
            node_table,
            fit_integrand,
            skip_this_job,
            max_number_cpu,
            master_process,
            fit_type_list,
            shared_lock,
            shared_event,
            shared_job_status,
            job_status_name,
         )



# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: University of Washington <https://www.washington.edu>
# SPDX-FileContributor: 2021-24 Bradley M. Bell
# ----------------------------------------------------------------------------
'''
{xrst_begin add_log_entry}
{xrst_spell
  unix
}

Add Log Table Entry
###################

Prototype
*********
{xrst_literal
   # BEGIN_DEF
   # END_DEF
}

Purpose
*******
Change the name of a table in the database.

connection
**********
is a dismod_at open connection to the database.

message
*******
is a ``str`` containing the message that is added at
the end of the log table

Log Table
*********
As row is added at the end of the log table with the
following columns values:

1. *log_id* : is the integer length of the log table before the message
2. *message_type* : is the text ``at_cascade``
3. *table_name* : is null
4. *row_id* : is null
5. *unix_time* : is the integer unit time
6. *message* : is the text message

{xrst_end   add_log_entry}
'''
import time
import dismod_at
# ----------------------------------------------------------------------------
# BEGIN_DEF
# at_cascade.add_log_entry
def add_log_entry(connection, message) :
   assert type(message) == str
   # END_DEF
   #
   # cmd
   cmd  = 'create table if not exists log('
   cmd += 'log_id       integer primary key,'
   cmd += 'message_type text,'
   cmd += 'table_name   text,'
   cmd += 'row_id       integer,'
   cmd += 'unix_time    integer,'
   cmd += 'message      text)'
   dismod_at.sql_command(connection, cmd)
   #
   # log_table
   log_table = dismod_at.get_table_dict(connection, 'log')
   #
   # seconds
   seconds   = int( time.time() )
   #
   # message_type
   message_type = 'at_cascade'
   #
   # cmd
   cmd = 'insert into log'
   cmd += ' (log_id,message_type,table_name,row_id,unix_time,message) values('
   cmd += str( len(log_table) ) + ','     # log_id
   cmd += f'"{message_type}",'            # message_type
   cmd += 'null,'                         # table_name
   cmd += 'null,'                         # row_id
   cmd += str(seconds) + ','              # unix_time
   cmd += f'"{message}")'                 # message
   dismod_at.sql_command(connection, cmd)



# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: University of Washington <https://www.washington.edu>
# SPDX-FileContributor: 2021-25 Bradley M. Bell
# ----------------------------------------------------------------------------
'''
{xrst_begin extract_avgint}

Create an Empty Directory
#########################

Prototype
*********
{xrst_literal ,
   # BEGIN_DEF ,    # END_DEF
   # BEGIN_RETURN , # END_RETURN
}

root_database
*************
On input this database contains a non-empty avgint table.
Upon return the avgint table in this database is empty.

avgint_table
************
This is the avgint table that was in the root_database on input.


{xrst_end extract_avgint}
'''
# -----------------------------------------------------------------------------
import dismod_at
import at_cascade
#
# BEGIN_DEF
# at_cascade.extract_avgint
def extract_avgint(root_database) :
   assert type(root_database) == str
   # END_DEF
   #
   # connection
   connection = dismod_at.create_connection(
      root_database, new = False, readonly = False
   )
   #
   # avgint_table
   avgint_table = dismod_at.get_table_dict(connection, 'avgint')
   #
   # root_database
   empty_table = list()
   message     = 'erase avgint table'
   tbl_name    = 'avgint'
   dismod_at.replace_table(connection, tbl_name, empty_table)
   at_cascade.add_log_entry(connection, message)
   #
   # connection
   connection.close()
   #
   # BEGIN_RETURN
   #
   assert type( avgint_table ) == list
   assert type( avgint_table[0] ) == dict
   return avgint_table
   # END_RETURN



# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: University of Washington <https://www.washington.edu>
# SPDX-FileContributor: 2021-25 Bradley M. Bell
# ----------------------------------------------------------------------------
r'''
{xrst_begin cascade_root_node}
{xrst_spell
  var
}

Cascade Fits Starting at Root Node
##################################

Prototype
*********
{xrst_literal
   # BEGIN_DEF
   # END_DEF
}

all_node_database
*****************
is a python string specifying the location of the
:ref:`all_node_db-name`
relative to the current working directory.

fit_goal_set
************
This is a ``set`` with elements of type ``int`` (``str``)
specifying the node_id (node_name) for each element of the
:ref:`glossary@fit_goal_set` .
Each such node must be the root node, or a descendant of the root node.
In addition, it must be in the :ref:`fit_goal_table-name` ,
or an ancestor of a node in the fit goal table.

no_ode_fit
**********
if ``True`` ( ``False`` ) the :ref:`no_ode_fit-name` routine
will (will not) be used to modify the mean of the parent value
and difference priors.

fit_type_list
*************
This is a list with one or two elements
and its possible elements are ``both`` and ``fixed``.
For each job, the first type of fit is attempted.
If it fails, and there is a second type of fit, it is attempted.
If it also fails, the corresponding job fails.

root_fit_database
*****************
This database is located at

   *result_dir*\ ``/``\ *root_node_name*\ ``/dismod.db``

see :ref:`option_all_table@result_dir`
and :ref:`option_all_table@root_node_name`.
If *no_ode_fit* is ``True`` ( ``False`` ) the priors in the
*root_database* are modified (are not modified)
before the root node is fit;
see :ref:`option_all_table@root_database` .
Upon return, this is a :ref:`glossary@fit_database` with the
extra properties listed below:

Version
=======
The ``at_cascade`` and ``dismod_at`` version numbers
are stored at the beginning of the log table in the
*root_fit_database* .

Output dismod.db
****************
Upon return for cascade_root_node,
the results for the fits are in ``dismod.db`` in the corresponding
directory relative to *result_dir*; i.e., sub-directory of *result_dir* .
The *.csv* files can be created using the
dismod_at db2csv command.
The dismod_at function ``plot_rate_fit`` and ``plot_data_fit``
can be used to crate the corresponding plots.

The root level fit the directory is
:ref:`option_all_table@root_node_name` .
If the current fit is just before a split,
there will be a sub-directory for the fit of each
:ref:`split_reference_table@split_reference_name` after the split.
Otherwise there will be a sub-directory for the fit of each child of the
node corresponding to the current fit.
You can determine the directory, relative to *result_dir*
corresponding to a fit using the :ref:`get_database_dir-name` function.


fit_var
=======
The fit_var table correspond to the posterior
mean for the model variables for  this job; i.e., this
:ref:`create_job_table@job_table@fit_node_id` and
:ref:`create_job_table@job_table@split_reference_id`.

sample
======
The sample table contains the corresponding samples from the posterior
distribution for the model variables for this job.

log
===
The log table contains a summary of the operations preformed on dismod.db
between it's input and output state.

{xrst_end cascade_root_node}
'''
# ----------------------------------------------------------------------------
import time
import os
import multiprocessing
import dismod_at
import at_cascade
# ----------------------------------------------------------------------------
# BEGIN_DEF
# at_cascade.cascade_root_node
def cascade_root_node(
   all_node_database       ,
   fit_goal_set            ,
   no_ode_fit              = False,
   fit_type_list           = [ 'both', 'fixed' ]
) :
   assert type(all_node_database)  == str
   assert type(fit_goal_set)       == set
   assert type(no_ode_fit)         == bool
   assert type(fit_type_list)      == list
   # END_DEF
   #
   # split_reference_table, option_all_table
   connection  = dismod_at.create_connection(
      all_node_database, new = False, readonly = True
   )
   split_reference_table = dismod_at.get_table_dict(
      connection, 'split_reference'
   )
   option_all_table = dismod_at.get_table_dict(connection, 'option_all')
   connection.close()
   #
   # option_all_dict
   option_all_dict = dict()
   for row in option_all_table :
      option_all_dict[ row['option_name'] ] = row['option_value']
   #
   # root_node_name, root_database, max_number_cpu, result_dir
   root_node_name     = option_all_dict['root_node_name']
   root_database      = option_all_dict['root_database']
   result_dir         = option_all_dict['result_dir']
   max_number_cpu     = 1
   if 'max_number_cpu' in option_all_dict :
      max_number_cpu = int( option_all_dict['max_number_cpu'] )
   #
   # check root_node_name
   parent_node_name = at_cascade.get_parent_node(root_database)
   if parent_node_name != root_node_name :
      msg  = f'{root_database} parent_node_name = {parent_node_name}\n'
      msg += f'{all_node_database} root_node_name = {root_node_name}'
      assert False, msg
   #
   # root_fit_database
   root_fit_database = f'{result_dir}/{root_node_name}/dismod.db'
   if not os.path.exists( f'{result_dir}/{root_node_name}' ) :
      os.makedirs( f'{result_dir}/{root_node_name}' )
   if not no_ode_fit :
      at_cascade.copy_root_db(root_database, root_fit_database)
      at_cascade.omega_constraint(all_node_database, root_fit_database)
   else :
      at_cascade.no_ode_fit(
         all_node_database  = all_node_database,
         root_database      = root_database,
         option_all_dict    = option_all_dict,
         fit_type           = fit_type_list[0],
      )
   #
   # node_table, covariate_table, fit_integrand
   fit_or_root = at_cascade.fit_or_root_class(
      root_fit_database, root_database
   )
   node_table      = fit_or_root.get_table('node')
   covariate_table = fit_or_root.get_table('covariate')
   avgint_table    = fit_or_root.get_table('avgint')
   if len(avgint_table) != 0 :
      msg = 'cascade_root_node: avgint table is not empty'
      assert False, msg
   fit_integrand = at_cascade.get_fit_integrand(fit_or_root)
   fit_or_root.close()
   #
   # root_node_id
   root_node_id = at_cascade.table_name2id(node_table, 'node', root_node_name)
   #
   # root_split_reference_id
   if len(split_reference_table) == 0 :
      root_split_reference_id = None
   else :
      cov_info = at_cascade.get_cov_info(
         option_all_table, covariate_table, split_reference_table
      )
      root_split_reference_id = cov_info['split_reference_id']
   #
   # job_table
   job_table = at_cascade.create_job_table(
      all_node_database          = all_node_database,
      node_table                 = node_table,
      start_node_id              = root_node_id,
      start_split_reference_id   = root_split_reference_id,
      fit_goal_set               = fit_goal_set,
   )
   #
   # start_job_id
   start_job_id = 0
   #
   # skip_start_job
   skip_start_job = False
   #
   # fit_one_process
   at_cascade.fit_parallel(
      job_table         = job_table ,
      start_job_id      = start_job_id,
      all_node_database = all_node_database,
      node_table        = node_table,
      fit_integrand     = fit_integrand,
      skip_start_job    = skip_start_job,
      max_number_cpu    = max_number_cpu,
      fit_type_list     = fit_type_list,
      shared_unique     = '',
   )



# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: University of Washington <https://www.washington.edu>
# SPDX-FileContributor: 2021-25 Bradley M. Bell
# ----------------------------------------------------------------------------
'''
{xrst_begin check_log}

Checks Logs For Warnings and Errors
###################################

Prototype
*********
{xrst_literal ,
   # BEGIN_DEF, # END_DEF
   # BEGIN_RETURN, # END_RETURN
}

Purpose
*******
Read the logs for a cascade and return all the messages of a certain type.
The databases are not modified (are opened in a read only fashion).

message_type
************
is  equal to ``error``, ``warning`` or ``at_cascade`` .
The corresponding messages are returned.

all_node_database
*****************
specifies the location of the
:ref:`all_node_db-name`
relative to the current working directory.
This argument can't be ``None``.

root_database
*************
specifies the location of the dismod_at
:ref:`glossary@root_database`.

job_table
*********
This is the :ref:`create_job_table@job_table` that we are checking
the log messages in.
Only jobs for which :ref:`create_job_table@job_table@prior_only` is false
are included; i.e., only jobs that correspond to fits.

start_job_id
************
This is the job that the log messages should start at.
If this is None, the first job in the job table is the start job.
Jobs before this job in the :ref:`create_job_table@job_table`
are not included.

max_job_depth
*************
This is the number of generations below the start job that are included;
see :ref:`job_descendant@Node Depth Versus Job Depth` .
If *max_job_depth* is None, all the jobs below the start job are included.
If *max_job_depth* is zero, only the start job is included.

message_dict
************
For each :ref:`create_job_table@job_table@job_name` in the job table
that is a key in *message_dict*.  The corresponding value

| *message_dict* [ *job_name* ]


is a non-empty ``list`` of ``str`` containing the messages for that job.
If a *job_name* is not a *key* is in *message_dict*,
there were no messages of the specified type for that job.

{xrst_end check_log}
'''
# ----------------------------------------------------------------------------
import time
import os
import multiprocessing
import dismod_at
import at_cascade
# ----------------------------------------------------------------------------
# BEGIN_DEF
# at_cascade.check_log
def check_log(
   message_type                  ,
   all_node_database             ,
   root_database            ,
   job_table                     ,
   start_job_id           = None ,
   max_job_depth          = None ,
) :
   assert type(message_type)        == str
   assert type(all_node_database)   == str
   assert type(root_database)  == str
   if start_job_id == None :
      start_job_id = 0
   assert max_job_depth == None or type(max_job_depth) == int
   # END_DEF
   #
   assert message_type in [ 'error', 'warning', 'at_cascade' ]
   #
   # node_table, covariate_table
   connection      = dismod_at.create_connection(
      root_database, new = False, readonly = True
   )
   node_table      = dismod_at.get_table_dict(connection, 'node')
   covariate_table = dismod_at.get_table_dict(connection, 'covariate')
   connection.close()
   #
   # split_reference_table, option_all_table, node_split_table
   connection  = dismod_at.create_connection(
      all_node_database, new = False, readonly = True
   )
   split_reference_table = dismod_at.get_table_dict(
      connection, 'split_reference'
   )
   option_all_table = dismod_at.get_table_dict(connection, 'option_all')
   node_split_table = dismod_at.get_table_dict(connection, 'node_split')
   connection.close()
   #
   # root_node_name
   result_dir     = None
   root_node_name = None
   for row in option_all_table :
      if row['option_name'] == 'result_dir' :
         result_dir = row['option_value']
      if row['option_name'] == 'root_node_name' :
         root_node_name = row['option_value']
   assert result_dir is not None
   assert root_node_name is not None
   #
   # check root_node_name
   parent_node_name = at_cascade.get_parent_node(root_database)
   if parent_node_name != root_node_name :
      msg  = f'{root_database} parent_node_name = {parent_node_name}\n'
      msg  = f'{all_node_database} root_node_name = {root_node_name}'
      assert False, msg
   #
   # root_node_id
   root_node_id = at_cascade.table_name2id(node_table, 'node', root_node_name)
   #
   if len(split_reference_table) == 0 :
      root_split_reference_id = None
   else :
      cov_info = at_cascade.get_cov_info(
         option_all_table, covariate_table, split_reference_table
      )
      root_split_reference_id = cov_info['split_reference_id']
   #
   # node_split_set
   node_split_set = set()
   for row in node_split_table :
      node_split_set.add( row['node_id'] )
   #
   # message_dict
   message_dict = dict()
   #
   # job_id
   for job_id in range( len(job_table) ) :
      #
      # include_this_job
      job_depth = at_cascade.job_descendant(job_table, start_job_id, job_id)
      if job_depth == None :
         include_this_job = False
      elif max_job_depth == None :
         include_this_job = True
      else :
         include_this_job = job_depth <= max_job_depth
      if include_this_job :
         include_this_job = not job_table[job_id]['prior_only']
      if include_this_job :
         #
         # job_name
         job_name = job_table[job_id]['job_name']
         #
         # fit_node_id
         fit_node_id = job_table[job_id]['fit_node_id']
         #
         # fit_split_reference_id
         fit_split_reference_id = job_table[job_id]['split_reference_id']
         #
         # fit_database
         database_dir = at_cascade.get_database_dir(
            node_table              = node_table,
            split_reference_table   = split_reference_table,
            node_split_set          = node_split_set,
            root_node_id            = root_node_id,
            root_split_reference_id = root_split_reference_id,
            fit_node_id             = fit_node_id ,
            fit_split_reference_id  = fit_split_reference_id,
         )
         fit_database      = f'{result_dir}/{database_dir}/dismod.db'
         #
         # log_table
         if not os.path.exists(fit_database) :
            message = f'Missing fit_database {fit_database}'
            message_dict[job_name] = [ message ]
         else :
            connection = dismod_at.create_connection(
                     fit_database, new = False, readonly = True
            )
            log_table  = dismod_at.get_table_dict(connection, 'log')
            connection.close()
            #
            # row
            for row in log_table :
               #
               if row['message_type'] == message_type :
                  #
                  # message_dict
                  if job_name not in message_dict :
                     message_dict[job_name] = list()
                  message_dict[job_name].append( row['message'] )
   #
   # BEGIN_RETURN
   # ...
   assert type(message_dict) == dict
   return message_dict
   # END_RETURN



# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: University of Washington <https://www.washington.edu>
# SPDX-FileContributor: 2021-24 Bradley M. Bell
# ----------------------------------------------------------------------------
'''
{xrst_begin omega_constraint}
{xrst_spell
  nslist
  smoothings
}

Set Omega Constraints in a Fit Node Database
############################################

Prototype
*********
{xrst_literal
   # BEGIN_DEF
   # END_DEF
}

all_node_database
*****************
is a python string containing the name of the :ref:`all_node_db-name`.

Use
===
This routine builds the *omega* constraints using the
:ref:`glossary@omega` data in the *all_node_database* .
If a node does not have *omega* data,
the data for the closest ancestor is used.
If a node does not have an ancestor with *omega* data,
zero is used for the *omega* constraint for that node.

fit_database
************
is a python string containing the name of a :ref:`glossary@fit_database`.

parent_node
===========
We use *parent_node* to refer to the parent node in the
dismod_at option table in the *fit_database*; i.e.,
the :ref:`glossary@fit_node` .

rate_table
==========
On input, the parent and child smoothing for omega must be null.
On return, they will be set to smoothings that yield the omega constraints.

smooth Table
============
Extra rows for the omega constraints will be added at the
end of the smooth table so that existing smoothings are preserved.

smooth_grid Table
=================
Extra rows for the omega constraints will be added at the
end of the smooth_grid table so existing smoothings are preserved.

nslist Table
============
On input, this table must be empty.
Upon return, it will contain entries that implement the omega constraints
for the children of the *parent_node*.

nslist_pair Table
=================
On input, this table must be empty.
Upon return, it will contain entries that implement the omega constraints
for the children of the *parent_node*.

Other Tables
============
None of the other tables in the database are modified.

{xrst_end omega_constraint}
'''
# ----------------------------------------------------------------------------
import copy
import dismod_at
import at_cascade
from math import log
# ----------------------------------------------------------------------------
def child_node_id_list(node_table, parent_node_id) :
   result = list()
   for (node_id, row) in enumerate(node_table) :
      if row['parent'] == parent_node_id :
         result.append(node_id)
   return result
# ----------------------------------------------------------------------------
# BEGIN_DEF
# at_cascade.omega_constraint
def omega_constraint(
   all_node_database ,
   fit_database ,
) :
   assert type(all_node_database) == str
   assert type(fit_database) == str
   # END_DEF
   #
   # all_tables
   connection        = dismod_at.create_connection(
      all_node_database, new = False, readonly = True
   )
   all_tables = dict()
   for name in [
      'option_all',
      'omega_all',
      'omega_index',
      'omega_age_grid',
      'omega_time_grid',
      'split_reference',
   ] :
      all_tables[name] = dismod_at.get_table_dict(connection, name)
   connection.close()
   #
   # case where omega constrained to zero
   if len( all_tables['omega_time_grid']) == 0 :
      assert len( all_tables['omega_all'] ) == 0
      assert len( all_tables['omega_index'] ) == 0
      assert len( all_tables['omega_age_grid'] ) == 0
      return
   #
   # n_omega_age, n_omega_time
   n_omega_age  = len( all_tables['omega_age_grid'] )
   n_omega_time = len( all_tables['omega_time_grid'] )
   #
   # root_database
   root_database      = None
   for row in all_tables['option_all'] :
      if row['option_name'] == 'root_database' :
         root_database      = row['option_value']
   assert root_database != None
   #
   # fit_tables
   fit_or_root = at_cascade.fit_or_root_class(
      fit_database, root_database
   )
   fit_tables   = dict()
   fit_null_row = dict()
   for name in [
      'age',
      'time',
      'covariate',
      'nslist',
      'nslist_pair',
      'node',
      'option',
      'rate',
      'smooth',
      'smooth_grid',
   ] :
      fit_tables[name]   = fit_or_root.get_table(name)
      fit_null_row[name] = fit_or_root.null_row(name)
   fit_or_root.close()
   #
   # check some fit_database assumptions
   assert len( fit_tables['nslist'] ) == 0
   assert len( fit_tables['nslist_pair'] ) == 0
   for row in fit_tables['rate'] :
      if row['rate_name'] == 'omega' :
         assert row['parent_smooth_id'] is None
         assert row['child_smooth_id'] is None
         assert row['child_nslist_id'] is None
   for row in all_tables['omega_age_grid'] :
      age_id = row['age_id']
      if age_id >= len( fit_tables['age'] ) :
         msg  = f'The age_id {age_id} not valid for the root_database'
         msg += f'\nbut it appears in the omega_age_grid table '
         msg += f'of the all_node_database'
         assert False, msg
   for row in all_tables['omega_time_grid'] :
      time_id = row['time_id']
      if time_id >= len( fit_tables['time'] ) :
         msg  = f'The time_id {time_id} not valid for the root_database'
         msg += f'\nbut it appears in the omega_time_grid table '
         msg += f'of the all_node_database'
         assert False, msg
   #
   # split_reference_id
   cov_info = at_cascade.get_cov_info(
      all_tables['option_all'],
      fit_tables['covariate'],
      all_tables['split_reference'],
   )
   if len( all_tables['split_reference'] )  == 0 :
      split_reference_id = None
   else :
      split_reference_id = cov_info['split_reference_id']
   #
   # parent_node_id
   parent_node_name = None
   for row in fit_tables['option'] :
      assert row['option_name'] != 'parent_node_id'
      if row['option_name'] == 'parent_node_name' :
         parent_node_name = row['option_value']
   assert parent_node_name is not None
   parent_node_id = at_cascade.table_name2id(
      fit_tables['node'], 'node', parent_node_name
   )
   #
   # node_id2omega_all_id
   node_id2omega_all_id = dict()
   for row in all_tables['omega_index'] :
      omega_all_id = row['omega_all_id']
      if omega_all_id % (n_omega_age * n_omega_time) != 0 :
         msg  = 'omega_index table: Expect omega_all_id to be a multiple '
         msg += 'of n_omega_age * n_omega_time\n'
         msg += f'omega_all_id = {omega_all_id} '
         msg += f'n_omega_age = {n_omega_age} '
         msg += f'n_omega_time = {n_omega_time} '
         assert False, msg
      if row['split_reference_id'] == split_reference_id :
         node_id2omega_all_id[ row['node_id'] ] = omega_all_id
   #
   # omega_ancestor_node_id
   node_id = parent_node_id
   while not node_id in node_id2omega_all_id :
      node_id = fit_tables['node'][node_id]['parent']
      if node_id is None :
         msg  = 'omega_constraint: no ancestor of ' + parent_node_name
         msg += ' has omega data'
         assert False, msg
   omega_ancestor_node_id = node_id
   assert not omega_ancestor_node_id is None
   #
   # parent_omega
   omega_all_id = node_id2omega_all_id[omega_ancestor_node_id]
   parent_omega   = list()
   for ij in range( n_omega_age * n_omega_time ) :
      row  = all_tables['omega_all'][omega_all_id + ij ]
      parent_omega.append( row['omega_all_value'] )
   #
   # parent_smooth_id
   parent_smooth_id  = len(fit_tables['smooth'])
   #
   # fit_tables['sooth_table']
   row           = copy.copy( fit_null_row['smooth'] )
   row['n_age']  = n_omega_age
   row['n_time'] = n_omega_time
   fit_tables['smooth'].append( row )
   #
   # fit_tables['smooth_grid']
   for i in range( n_omega_age ) :
      for j in range( n_omega_time ) :
         row     = copy.copy( fit_null_row['smooth_grid'] )
         age_id  = all_tables['omega_age_grid'][i]['age_id']
         time_id = all_tables['omega_time_grid'][j]['time_id']
         ij      = i * n_omega_time + j
         omega   = parent_omega[ij]
         row['age_id']      = age_id
         row['time_id']     = time_id
         row['smooth_id']   = parent_smooth_id
         row['const_value'] = omega
         fit_tables['smooth_grid'].append( row )
   #
   # child_node_list
   child_node_list = child_node_id_list(fit_tables['node'], parent_node_id)
   #
   # nslist_id
   nslist_id = len( fit_tables['nslist'] )
   #
   # child_node_id
   for child_node_id in child_node_list :
      #
      # child_omega
      if not child_node_id in node_id2omega_all_id :
         child_omega = parent_omega
      else :
         omega_all_id = node_id2omega_all_id[child_node_id]
         child_omega  = list()
         for ij in range( n_omega_age * n_omega_time ) :
               row  = all_tables['omega_all'][omega_all_id + ij ]
               child_omega.append( row['omega_all_value'] )
      #
      # random_effect
      random_effect = list()
      for ij in range( n_omega_age * n_omega_time ) :
         if parent_omega[ij] <= 0 :
            msg  = 'parent_omega <= 0\n'
            msg += f'parent_node_id = {parent_node_id}'
            msg += f', omega_ancestor_node_id = {omega_ancestor_node_id}'
            msg += f', parent_omega = {parent_omega[ij]}'
            assert False, msg
         if child_omega[ij] <= 0 :
            msg  = 'child_omega <= 0'
            msg += f', child_node_id = {child_node_id}'
            if child_node_id in node_id2omega_all_id :
               msg += f'\nomega_ancestor_node_id = {child_node_id}'
            else :
               msg += '\nomega_ancestor_node_id = '
               msg += str(omega_ancestor_node_id)
            msg += f', child_omega = {child_omega[ij]}'
            assert False, msg
         random_effect.append( log( child_omega[ij] / parent_omega[ij] ) )
      #
      # smooth_id
      smooth_id = len( fit_tables['smooth'] )
      #
      # fit_tables['nslist_pair']
      row              = copy.copy( fit_null_row['nslist_pair'] )
      row['nslist_id'] = nslist_id
      row['node_id']   = child_node_id
      row['smooth_id'] = smooth_id
      fit_tables['nslist_pair'].append( row )
      #
      # fit_tables['smooth']
      row           = copy.copy( fit_null_row['smooth'] )
      row['n_age']  = n_omega_age
      row['n_time'] = n_omega_time
      fit_tables['smooth'].append( row )
      #
      # fit_tables['smooth_grid']
      for i in range( n_omega_age ) :
         for j in range( n_omega_time ) :
            row     = copy.copy( fit_null_row['smooth_grid'] )
            age_id  = all_tables['omega_age_grid'][i]['age_id']
            time_id = all_tables['omega_time_grid'][j]['time_id']
            row['age_id']      = age_id
            row['time_id']     = time_id
            row['smooth_id']   = smooth_id
            row['const_value'] = random_effect[i * n_omega_time + j]
            fit_tables['smooth_grid'].append( row )
   #
   # fit_tables['nslist']
   row                = copy.copy( fit_null_row['nslist'] )
   row['nslist_name'] = 'child_omega'
   fit_tables['nslist'].append( row )
   #
   # fit_tables['rate']
   for row in fit_tables['rate'] :
      if row['rate_name'] == 'omega' :
         row['parent_smooth_id'] = parent_smooth_id
         row['child_nslist_id']  = nslist_id
   #
   # replace these fit tables
   connection    = dismod_at.create_connection(
      fit_database, new = False, readonly = False
   )
   #
   for name in [
      'nslist',
      'nslist_pair',
      'option',
      'rate',
      'smooth',
      'smooth_grid',
   ] :
      dismod_at.replace_table(connection, name, fit_tables[name])
   #
   connection.close()



# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: University of Washington <https://www.washington.edu>
# SPDX-FileContributor: 2021-24 Bradley M. Bell
# ----------------------------------------------------------------------------
import at_cascade
'''
{xrst_begin csv.covariate_spline}
{xrst_spell
  cov
}

Create Spline Functions For Covariates
######################################

Prototype
*********
{xrst_literal ,
   BEGIN_DEF, END_DEF
   BEGIN_RETURN, END_RETURN
}

covariate_table
***************
Is a ``list`` of ``dict`` representation of a
:ref:`csv.simulate@Input Files@covariate.csv` file.
All of the keys in this ``dict`` are covariate names except for
``node_name`` , ``sex``, ``age`` , ``time`` and ``omega`` .
All of the columns have been converted to type ``float`` except
for *node_name* and *sex* which have type ``str`` .

node_set
********
This is the set of nodes in :ref:`csv.simulate@Input Files@node.csv`

age_grid
********
This is a sorted list of float containing the age values
that are present for each node_name and sex.

time_grid
*********
This is a sorted list of float containing the time values
that are present for each node_name and sex.

spline_cov
**********
This is a dict of dict of dict.
Let spline be defined by

   for node_name in covariate_table
      for sex in covariate_table
         for cov_name a covariate name or omega
            spline = spline_cov[node_name][sex][cov_name]

Then spline is the corresponding function of age and time; i.e.,
{xrst_code py}
   z = spline(age, time)
{xrst_code}
sets *z* to the value of the covariate or omega
for the specified age, time, node_name, and sex where
*age*, *time*, and *z* are floats.

Side Effects
************
This routine aborts with an assert if covariate.csv does not have the
same rectangular grid in age and time for each (node_name, sex) pair.

{xrst_end csv.covariate_spline}
'''
# BEGIN_DEF
# at_cascade.csv.covariate_spline
def covariate_spline(covariate_table , node_set) :
   assert type( covariate_table ) == list
   assert type( covariate_table[0] ) == dict
   assert type( node_set ) == set
   # END_DEF
   #
   # cov_name_list
   exclude       = {  'node_name', 'sex', 'age', 'time' }
   cov_name_set  = set( covariate_table[0].keys() ) - exclude
   cov_name_list = list( cov_name_set )
   #
   # covariate_row_list
   covariate_row_list  = dict()
   line_number         = 0
   for row in covariate_table :
      line_number += 1
      node_name    = row['node_name']
      sex          = row['sex']
      age          = float( row['age'] )
      time         = float( row['time'] )
      if node_name not in node_set :
         msg  = f'covariate_spline: Error: '
         msg += f'line {line_number} in covariate.csv\n'
         msg += f'node_name {node_name} is not in node.csv'
         assert False, msg
      if sex not in ['female', 'male', 'both'] :
         msg  = f'covariate_spline: Error: '
         msg += f'line {line_number} in covariate.csv\n'
         msg += f'sex {sex} is not female or male'
         assert False, msg
      #
      # covariate_row_list
      if node_name not in covariate_row_list :
         covariate_row_list[node_name] = dict()
      if sex not in covariate_row_list[node_name] :
         covariate_row_list[node_name][sex] = list()
      covariate_row_list[node_name][sex].append( row )
   #
   # cov_same
   cov_same = at_cascade.csv.covariate_same(covariate_table)
   #
   # spline_cov
   spline_cov = dict()
   previous   = dict()
   for node_name in covariate_row_list :
      #
      # spline_cov[node_name]
      spline_cov[node_name] = dict()
      #
      for sex in covariate_row_list[node_name] :
         #
         # spline_cov[node_name][sex]
         spline_cov[node_name][sex] = dict()
         #
         # z_list
         # only compuite splines that are needed
         z_list = list()
         for cov_name in cov_name_list :
            triple = (node_name, sex, cov_name)
            if cov_same[triple] == triple :
               z_list.append( cov_name )
         if 0 < len(z_list) :
            #
            # age_grid, time_grid, spline_dict
            age_grid, time_grid, spline_dict = at_cascade.bilinear(
               table  = covariate_row_list[node_name][sex] ,
               x_name = 'age',
               y_name = 'time',
               z_list = cov_name_list
            )
            # covariate_same should have checked rectangular grid
            assert spline_dict != None
            if len( previous ) != 0 :
               same_grid = previous['age_grid'] == age_grid
               same_grid = same_grid and previous['time_grid'] == time_grid
               assert same_grid
            #
            # spline_cov[node_name][sex]
            for cov_name in z_list :
               triple = (node_name, sex, cov_name)
               spline_cov[node_name][sex][cov_name] = spline_dict[cov_name]
            #
            previous['node_name'] = node_name
            previous['sex']       = sex
            previous['age_grid']  = age_grid
            previous['time_grid'] = time_grid
   #
   # spline_cov
   # link from splines that are not needed to corresponding same spline
   for node_name in covariate_row_list :
      for sex in covariate_row_list[node_name] :
         for cov_name in cov_name_list :
            triple = (node_name, sex, cov_name)
            if cov_same[triple] != triple :
               (node_other, sex_other, cov_other) = cov_same[triple]
               assert cov_other == cov_name
               spline_cov[node_name][sex][cov_name] = \
                  spline_cov[node_other][sex_other][cov_other]

   # BEGIN_RETURN
   # ...
   assert type(age_grid) == list
   assert type(time_grid) == list
   assert type(spline_cov) == dict
   return age_grid, time_grid, spline_cov
   # END_RETURN



# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: University of Washington <https://www.washington.edu>
# SPDX-FileContributor: 2021-24 Bradley M. Bell
# ----------------------------------------------------------------------------
'''
{xrst_begin get_cov_info}
{xrst_spell
  rel
}

Get Covariate Information
#########################

Prototype
*********
{xrst_literal ,
   # BEGIN_DEF, # END_DEF
   # BEGIN_RETURN, # END_RETURN
}

option_all_table
****************
This is the :ref:`option_all_table-name` as a python list
of python dictionaries.
This argument can't be ``None``.

covariate_table
***************
This is the dismod_at covariate table as a python list
of python dictionaries.
This argument can't be ``None``.

split_reference_table
*********************
This is the :ref:`split_reference_table-name` as a python list
of python dictionaries.
This argument can't be ``None``.

cov_info
********
The return value *cov_info* is a `dict` with the following keys:

abs_covariate_id_set
====================
if *key* is abs_covariate_id_set, *cov_info[key]* is a ``set`` of ``int``.
A covariate_id is in this set if and only if the corresponding
covariate name is in :ref:`option_all_table@absolute_covariates`.
If absolute_covariates does not appear in the option_all table,
*cov_info[key]* is the empty set.
The reference value for these absolute covariates is the same
for all nodes and all values of the splitting covariate.

rel_covariate_id_set
====================
if *key* is rel_covariate_id_set, *cov_info[key]* is a set of ``int``.
This is the set of covariate_id values corresponding to the
:ref:`relative covariates<glossary@Relative Covariate>`.

split_covariate_id
==================
If :ref:`split_reference_table-name` is empty, this key is not present.
Otherwise, for *key* equal to split_covariate_id, *cov_info[key]* is an ``int``
equal to the covariate_id corresponding to the
:ref:`option_all_table@split_covariate_name`.

split_reference_list
====================
If :ref:`split_reference_table-name` is empty, this key is not present.
Otherwise, for *key* equal to split_reference_list, *cov_info[key]* is a
``list`` of ``float`` representation of
:ref:`split_reference_table@split_reference_value` in the
same order as they appear in the table.

split_reference_id
==================
If :ref:`split_reference_table-name` is empty, this key is not present.
Otherwise, for *key* equal to split_reference_id, *cov_info[key]* is an ``int``
containing an index in the split_reference_list.
The corresponding value in split_reference_list is equal to
the reference value for split_covariate_name in the covariate table.

{xrst_end get_cov_info}
'''
import at_cascade
# BEGIN_DEF
# at_cascade.get_cov_info
def get_cov_info(
   option_all_table      ,
   covariate_table       ,
   split_reference_table ,
) :
   assert type(option_all_table) == list
   assert type(covariate_table) == list
   assert type(split_reference_table) == list
   # END_DEF
   #
   # option_all
   option_all = dict()
   for row in option_all_table :
      option_all[ row['option_name'] ] = row['option_value']
   #
   # check split_covariate_name
   if len(split_reference_table) == 0 :
      assert not 'split_covariate_name' in option_all
   else :
      assert 'split_covariate_name' in option_all
   #
   # split_reference_list
   split_reference_list = list()
   for row in split_reference_table :
      split_reference_list.append( row['split_reference_value'] )
   #
   # abs_covariate_id_set
   abs_covariate_id_set = set()
   if 'absolute_covariates' in option_all :
      absolute_covariate_list = option_all['absolute_covariates'].split()
      for covariate_name in absolute_covariate_list :
         covariate_id = at_cascade.table_name2id(
            covariate_table, 'covariate', covariate_name
         )
         abs_covariate_id_set.add(covariate_id)
   #
   # rel_covariate_id_set
   rel_covariate_id_set = set()
   for covariate_id in range( len(covariate_table) ) :
      if not covariate_id in abs_covariate_id_set :
         rel_covariate_id_set.add( covariate_id )
   #
   # case where no splitting covariate
   if len(split_reference_table) == 0 :
      cov_info = {
         'abs_covariate_id_set':  abs_covariate_id_set,
         'rel_covariate_id_set':  rel_covariate_id_set,
      }
      return cov_info
   #
   # split_covarate_name
   split_covariate_name = option_all['split_covariate_name']
   #
   # split_covariate_id
   split_covariate_id   = at_cascade.table_name2id(
      covariate_table, 'covariate', split_covariate_name
   )
   #
   # rel_covariate_id_set
   rel_covariate_id_set.discard( split_covariate_id )
   #
   # split_reference
   split_reference = covariate_table[split_covariate_id]['reference']
   #
   # split_reference_id
   if not split_reference in split_reference_list :
      msg  = 'Cannot find covariate table value for splitting covariate '
      msg += 'in split_reference table\n'
      msg += f'split_reference_value = {split_reference_list}, '
      msg += f'split_covariate_id = {split_covariate_id}, '
      msg += f'covariate table reference value = {split_reference}'
      assert False, msg
   split_reference_id = split_reference_list.index( split_reference )
   #
   cov_info = {
      'abs_covariate_id_set':  abs_covariate_id_set,
      'rel_covariate_id_set':  rel_covariate_id_set,
      'split_covariate_id':    split_covariate_id,
      'split_reference_list':  split_reference_list,
      'split_reference_id':    split_reference_id,
   }
   # BEGIN_RETURN
   # ...
   assert type(cov_info) == dict
   return cov_info
   # END_RETURN



# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: University of Washington <https://www.washington.edu>
# SPDX-FileContributor: 2021-24 Bradley M. Bell
# ----------------------------------------------------------------------------
'''
{xrst_begin clear_shared}
{xrst_spell
  errno
  inuse
}

Clear at_cascade Shared Memory
##############################

Prototype
*********
{xrst_literal
   # BEGIN_DEF
   # END_DEF
}

Purpose
*******
This routine clears the at_cascade shared memory that is in use
and has a particular prefix.
For example, you may get the following error message when trying
to run the cascade:

   FileExistsError: [Errno 17] File exists: *name*

where *name* ends with ``_number_cpu_inuse`` or ``_job_status``.
This may happen if the previous :ref:`fit_parallel-name`
did not terminate cleanly; e.g., if the system crashed.

all_node_database
*****************
is a path to the :ref:`all_node_db-name`.
This is used to determine the
:ref:`option_all_table@shared_memory_prefix`.

job_name
********
This is the :ref:`create_job_table@job_table@job_name`
for the shared memory that we are clearing; see
:ref:`fit_parallel@shared_unique` .

{xrst_end clear_shared}
'''
import dismod_at
import at_cascade
from multiprocessing import shared_memory
# BEGIN_DEF
# at_cascade.clear_shared
def clear_shared(all_node_database, job_name) :
   assert type(all_node_database) == str
   assert type(job_name) == str
   # END_DEF
   #
   # shared_memory_prefix
   connection           = dismod_at.create_connection(
      all_node_database, new = False, readonly = True
   )
   option_all_table     = dismod_at.get_table_dict(connection, 'option_all')
   connection.close()
   shared_memory_prefix = ""
   for row in option_all_table :
      if row['option_name'] == 'shared_memory_prefix' :
         shared_memory_prefix = row['option_value']
   #
   #
   # shared_memory_prefix_plus
   shared_memory_prefix_plus = f'{shared_memory_prefix}_{job_name}'
   #
   # name
   for name in  [ '_number_cpu_inuse', '_job_status' ] :
      #
      # shared_memory_name
      shared_memory_name = shared_memory_prefix_plus + name
      mapped_memory_name = at_cascade.map_shared( shared_memory_name )
      try :
         shm = shared_memory.SharedMemory(
            create = True, name = mapped_memory_name, size = 1
         )
         exists = False
      except FileExistsError :
         try :
            shm = shared_memory.SharedMemory(
               create = False, name = mapped_memory_name
            )
            exists = True
         except Exception as e :
            print( str(e) )
            exists = None
      #
      if exists == None :
         print( f'Unknown state: {shared_memory_name}' )
         print( 'Try re-running clear_shared')
      else :
         if exists :
            print( f'Found:     {shared_memory_name}' )
         else :
            print( f'Not Found: {shared_memory_name}' )
         if shared_memory_name != mapped_memory_name :
            print( f'           {mapped_memory_name}' )
         #
         shm.close()
         shm.unlink()
   print('clear_shared: OK')



# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: University of Washington <https://www.washington.edu>
# SPDX-FileContributor: 2021-25 Bradley M. Bell
# ----------------------------------------------------------------------------
import os
import copy
import math
import shutil
import at_cascade
import dismod_at
#
"""
{xrst_begin csv.set_truth}
{xrst_spell
  var
}

Set the Truth Table for a Fitted Database
#########################################

Prototype
*********
{xrst_literal
   # BEGIN_SET_TRUTH
   # END_SET_TRUTH
}

sim_dir
*******
Is the :ref:`csv.simulate@sim_dir` for the simulation
of this cascade data set.

fit_database
************
Is the name of a database, relative to the current directory,
that corresponds to a fit.

#. A new ``truth_var`` table is written to this database
   containing the dismod_at variable values corresponding to truth
   during the simulation.

#. Note that the rate random effects are in log space and relative
   to the fixed effect rate at the same age and time.

#. The contents of the database, and not its location relative to the
   :ref:`csv.fit@fit_dir` is used to determine the node and sex for
   the truth. In fact, this routine does not know what the *fit_dir* is.

root_database
*************
is the :ref:`glossary@root_database` .

{xrst_end csv.set_truth}
"""
# BEGIN_SET_TRUTH
# at_cascade.csv.set_truth
def set_truth(sim_dir, fit_database, root_database) :
   assert type(sim_dir) == str
   assert type(fit_database) == str
   assert type(root_database) == str
   # END_SET_TRUTH
   #
   # fit_node_dir
   index        = fit_database.rfind('/')
   fit_node_dir = fit_database[: index]
   #
   # fit_table
   fit_or_root = at_cascade.fit_or_root_class(
      fit_database, root_database
   )
   fit_table = dict()
   for table_name in [
      'age', 'covariate', 'integrand', 'node', 'rate', 'time', 'var'
   ] :
      fit_table[table_name] = fit_or_root.get_table(table_name)
   fit_or_root.close()
   #
   # fit_node_name
   fit_node_name = at_cascade.get_parent_node(fit_database)
   fit_node_id   = at_cascade.table_name2id(
      table    = fit_table['node'] ,
      tbl_name = 'node'            ,
      row_name = fit_node_name     ,
   )
   #
   # cov2reference
   cov2reference = dict()
   for row in fit_table['covariate'] :
      cov2reference[ row['covariate_name'] ] = float( row['reference'] )
   assert 'sex' in cov2reference
   #
   # sex_name
   sex_name = None
   for key in at_cascade.csv.sex_name2value :
      if at_cascade.csv.sex_name2value[key] == cov2reference['sex'] :
         sex_name = key
   assert sex_name != None
   #
   # rate2integrand
   rate2integrand = {
      'pini'  : 'prevalence' ,
      'iota'  : 'Sincidence' ,
      'rho'   : 'remission'  ,
      'chi'   : 'mtexcess'   ,
      'omega' : 'mtother'    ,
   }
   #
   # simulate_table, var_id2simulate_id
   simulate_table        = list()
   var_id2simulate_id    = list()
   simulate_id           = -1
   for var_row in fit_table['var'] :
      #
      # var_type
      var_type  = var_row['var_type']
      assert not var_type.startswith('mulstd_')
      assert not var_type.startswith('mulcov_meas_')
      #
      if var_type != 'rate' :
         var_id2simulate_id.append( None )
      else :
         # node_name, age, time, rate_name
         node_name  = fit_table['node'][  var_row['node_id'] ]['node_name']
         age        = fit_table['age'][  var_row['age_id']  ]['age']
         time       = fit_table['time'][ var_row['time_id'] ]['time']
         rate_name  = fit_table['rate'][ var_row['rate_id'] ]['rate_name']
         #
         # age
         if rate_name == 'pini' :
            age = fit_table['age'][0]['age']
            # This is really OK, just want a heads up in this case
            assert age == 0.0
         #
         # simulate_id, simulate_table, var_id2simulate_id
         simulate_id += 1
         integrand_name = rate2integrand[ rate_name]
         sim_row = {
            'simulate_id'    : simulate_id  ,
            'integrand_name' : integrand_name ,
            'node_name'      : fit_node_name  ,
            'sex'            : sex_name       ,
            'age_lower'      : age            ,
            'age_upper'      : age            ,
            'time_lower'     : time           ,
            'time_upper'     : time           ,
            'meas_std_cv'    : 0.0            ,
            'meas_std_min'   : 1.0            ,
         }
         simulate_table.append( copy.copy(sim_row ) )
         var_id2simulate_id.append( simulate_id )
         #
         if node_name != fit_node_name :
            # simulate_table
            # This is a random effects hence its rate value is in log space
            # and is relative to the corresponding fixed effect.
            simulate_id         += 1
            sim_row['simulate_id'] = simulate_id
            sim_row['node_name']   = node_name
            simulate_table.append( copy.copy(sim_row ) )
   #
   assert len( fit_table['var'] ) == len( var_id2simulate_id )
   #
   # copy_list
   # must copy these files because simulate may be run in parallel.
   copy_list = [
      'node', 'covariate', 'no_effect_rate', 'multiplier_sim', 'random_effect'
   ]
   #
   # copy files
   for name in copy_list :
      src = f'{sim_dir}/{name}.csv'
      dst = f'{fit_node_dir}/{name}.csv'
      shutil.copyfile(src, dst)
   #
   # option_sim.csv
   file_name        = f'{sim_dir}/option_sim.csv'
   option_sim_table = at_cascade.csv.read_table(file_name)
   for name in [ 'new_random_effects', 'trace' ] :
      found            = False
      for row in option_sim_table :
         if row['name'] == name :
            found = True
            row['value'] = 'false'
      if not found :
         row = { 'name' : name , 'value' : 'false' }
      option_sim_table.append( row )
   file_name  = f'{fit_node_dir}/option_sim.csv'
   at_cascade.csv.write_table(file_name, option_sim_table)
   #
   # covariate.csv
   file_name        = f'{fit_node_dir}/covariate.csv'
   covariate_table  = at_cascade.csv.read_table(file_name)
   for row in covariate_table :
      for cov in cov2reference :
         if cov != 'sex' and cov in row :
            row[cov] = cov2reference[cov]
   at_cascade.csv.write_table(file_name, covariate_table)
   #
   # simulate.csv
   at_cascade.csv.write_table(
      file_name = f'{fit_node_dir}/simulate.csv' ,
      table     = simulate_table
   )
   #
   # data_sim.csv, multiplier_sim.csv
   at_cascade.csv.simulate(fit_node_dir)
   #
   # sim_table
   sim_table = dict()
   for table_name in [ 'data_sim', 'multiplier_sim' ] :
      file_name             = f'{fit_node_dir}/{table_name}.csv'
      sim_table[table_name] = at_cascade.csv.read_table(file_name)
   #
   # remove copies
   if False :
      for name in copy_list :
         file_copy = f'{fit_node_dir}/{name}.csv'
         os.remove(file_copy)
   #
   # turth_var_table
   truth_var_table = list()
   covariate_set   = dict()
   for (var_id, var_row) in enumerate( fit_table['var'] ) :
      #
      # truth_var_value
      truth_var_value = None
      #
      # var_type
      var_type     = var_row['var_type']
      #
      # simulate_id
      simulate_id   = var_id2simulate_id[var_id]
      if simulate_id != None :
         assert var_type == 'rate'
         node_name   = fit_table['node'][  var_row['node_id'] ]['node_name']
         fixed_value = float( sim_table['data_sim'][simulate_id]['meas_mean'] )
         if node_name == fit_node_name :
            truth_var_value = fixed_value
         elif fixed_value == 0.0 :
            truth_var_value = 0.0
         else :
            random_value = \
               float( sim_table['data_sim'][simulate_id + 1]['meas_mean'] )
            truth_var_value = math.log( random_value / fixed_value )
      else :
         #
         # age_id, time_id
         age_id       = var_row['age_id']
         rate_id      = var_row['rate_id']
         time_id      = var_row['time_id']
         covariate_id = var_row['covariate_id']
         #
         if var_type == 'mulcov_rate_value' :
            #
            # rate_name, covariate_name
            rate_name      = fit_table['rate'][rate_id]['rate_name']
            covariate_name = \
               fit_table['covariate'][covariate_id]['covariate_name']
            #
            # truth_var_value
            for row in sim_table['multiplier_sim'] :
               if row['rate_name'] == rate_name :
                  if row['covariate_or_sex'] == covariate_name :
                     truth_var_value = row['multiplier_truth']
            assert truth_var_value != None
            #
            # covariate_set
            if rate_name not in covariate_set :
               covariate_set[rate_name] = set()
            assert covariate_name not in covariate_set[rate_name]
            covariate_set[rate_name].add( covariate_name )
         else :
            # This var_type not yet implemented
            assert False
      assert truth_var_value != None
      truth_var_table.append( { 'truth_var_value' : truth_var_value} )
   #
   # fit_database, connection
   # add the truth_var table
   connection = dismod_at.create_connection(
      fit_database, new = False, readonly = False
   )
   if at_cascade.table_exists(connection, table_name='truth_var') :
      dismod_at.replace_table(
         connection, tbl_name = 'truth_var', table_dict = truth_var_table
      )
   else :
      row_list = list()
      for row in truth_var_table :
         row_list.append( [ row['truth_var_value'] ] )
      dismod_at.create_table(
         connection,
         tbl_name = 'truth_var',
         col_name = [ 'truth_var_value' ],
         col_type = [ 'real' ],
         row_list = row_list,
      )
   #
   # fit_database, connection
   msg = 'csv.predict: setting truth_var table using simulation information'
   at_cascade.add_log_entry(connection, msg)
   connection.close()



# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: University of Washington <https://www.washington.edu>
# SPDX-FileContributor: 2021-25 Bradley M. Bell
# ----------------------------------------------------------------------------
# Set this to False when debugging an exception during pre_one_job routine
catch_exceptions_and_continue = True
# ----------------------------------------------------------------------------
r'''
{xrst_begin csv.pre_one_process}
{xrst_spell
  dtype
  numpy
}

Predict Using One Process
#########################
Keep predicting for as long as thee are jobs left to predict for.

Prototype
*********
{xrst_literal
   # BEGIN_DEF
   # END_DEF
}

fit_dir
*******
Same as the csv fit :ref:`csv.fit@fit_dir` .

sim_dir
*******
Same as :ref:`csv.predict@sim_dir` .

option_predict
**************
This is an in memory representation of
:ref:`csv.predict@Input Files@option_predict.csv` .

all_node_database
*****************
This is the all node database for this fit.

all_covariate_table
*******************
This is an in memory representation of
:ref:`csv.fit@Input Files@covariate.csv` .

job_table
*********
is the :ref:`create_job_table@job_table` for this cascade.

node_table
**********
is the dismod_at node table for this cascade.

root_node_id
************
is the node table id for the root node of the cascade.

root_split_reference_id
***********************
is the split_reference table id for the root job of the cascade

at_cascade_log_dict
*******************
For each :ref:`create_job_table@job_table@job_name` in the job table
that is a key in *at_cascade_log_dict*.  The corresponding value

| *at_cascade_log_dict* [ *job_name* ]


is a non-empty ``list`` of ``str`` containing ``at_cascade`` messages
in the log table for that job.
If a *job_name* is not a *key* is in *at_cascade_log_dict*,
there were no at_cascade messages for that job.

job_status_name
***************
is the name corresponding to each possible job status integer values.
If *i* is an integer job status value, *job_status_name* [ *i* ] is the
corresponding name.

.. csv-table::
   :header-rows: 1

   Name,    Meaning
   'skip',  job is not included in the predictions
   'ready', job is ready to run
   'run',   job is running
   'done',  job finished running

shared_job_status_name
**********************
This the name of the shared job status memory.
The corresponding multiprocessing shared memory is
a numpy array with ``dtype`` equal to ``int`` and
with length equal to the length of *job_table* .
The value *shared_job_status* [ *job_table_index* ] is the
integer status code for the corresponding job; see *job_status_name* above.

shared_lock
***********
This lock must be acquired during the time that
a process reads or changes *shared_job_status* .

Csv Output Files
****************
see :ref:`csv.pre_one_job@Csv Output Files`

Parallel Processing
*******************
Always copy ``dismod_at.db`` to another file before predicting with it
because ``dismod_at.db`` may be an ancestor for another prediction or fit
that is running in parallel.

{xrst_end csv.pre_one_process}
'''
# ----------------------------------------------------------------------------
import numpy
import os
import datetime
import shutil
import dismod_at
import at_cascade
import multiprocessing
# ----------------------------------------------------------------------------
# acquire_lock
def acquire_lock(shared_lock) :
   assert type(shared_lock) == multiprocessing.synchronize.Lock
   seconds = 10
   ok = shared_lock.acquire(
      block   = True,
      timeout = seconds
   )
   if not ok :
      msg = f'pre_one_process: did not obtain lock in {seconds} seconds'
      assert False, msg
# ----------------------------------------------------------------------------
# print_begin
def print_begin(job_name) :
   assert type(job_name) == str
   now         = datetime.datetime.now()
   str_time    = now.strftime("%H:%M:%S")
   msg  = f'Begin: {str_time}: predict {job_name}'
   print(msg)
# ----------------------------------------------------------------------------
# print_end
def print_end(
   job_name            ,
   n_done      = None  ,
   n_job_total = None  ,
   job_error   = None  ,
) :
   assert type(job_name) == str
   assert type(n_done)      == int
   assert type(n_job_total) == int
   assert type(job_error)   == str or job_error == None
   #
   now         = datetime.datetime.now()
   str_time    = now.strftime("%H:%M:%S")
   if job_error == None :
      msg  = f'End:   {str_time}: predict {job_name} {n_done}/{n_job_total}'
   else :
      msg  = f'Error: {str_time}: predict {job_name} {n_done}/{n_job_total}: '
      msg += job_error
   print(msg)
# ----------------------------------------------------------------------------
def try_one_job(
   predict_job_name        ,
   fit_dir                 ,
   sim_dir                 ,
   pre_database            ,
   predict_node_id         ,
   predict_sex_id          ,
   all_node_database       ,
   all_covariate_table     ,
   float_precision         ,
   fit_same_as_predict     ,
   option_predict          ,
) :
   assert type(predict_job_name) == str
   assert type(fit_dir) == str
   assert sim_dir == None or type(sim_dir) == str
   assert type(pre_database) == str
   assert type(predict_node_id) == int
   assert type(all_node_database) == str
   assert type(all_covariate_table) == list
   assert type( all_covariate_table[0] ) == dict
   assert type( float_precision ) == int
   assert type( fit_same_as_predict ) == bool
   assert type( option_predict ) == dict
   #
   # predict_job_error
   predict_job_error = None
   #
   if not catch_exceptions_and_continue :
      at_cascade.csv.pre_one_job(
         predict_job_name        = predict_job_name          ,
         fit_dir                 = fit_dir                   ,
         sim_dir                 = sim_dir                   ,
         pre_database            = pre_database              ,
         predict_node_id         = predict_node_id           ,
         predict_sex_id          = predict_sex_id            ,
         all_node_database       = all_node_database         ,
         all_covariate_table     = all_covariate_table       ,
         float_precision         = float_precision           ,
         fit_same_as_predict     = fit_same_as_predict       ,
         option_predict          = option_predict            ,
      )
   else :
      try :
         at_cascade.csv.pre_one_job(
            predict_job_name        = predict_job_name          ,
            fit_dir                 = fit_dir                   ,
            sim_dir                 = sim_dir                   ,
            pre_database            = pre_database              ,
            predict_node_id         = predict_node_id           ,
            predict_sex_id          = predict_sex_id            ,
            all_node_database       = all_node_database         ,
            all_covariate_table     = all_covariate_table       ,
            float_precision         = float_precision           ,
            fit_same_as_predict     = fit_same_as_predict       ,
            option_predict          = option_predict            ,
         )
         #
         # predict_job_error
         predict_job_error = None
      except Exception as e :
         predict_job_error = str(e)
   #
   assert predict_job_error == None or type(predict_job_error) == str
   return predict_job_error
# ----------------------------------------------------------------------------
# BEGIN_DEF
# at_cascade.csv.pre_one_process
def pre_one_process(
  fit_dir,
   sim_dir,
   option_predict,
   all_node_database,
   all_covariate_table,
   job_table,
   node_table,
   root_node_id,
   root_split_reference_id,
   at_cascade_log_dict,
   job_status_name,
   shared_job_status_name,
   shared_lock,
) :
   assert type(fit_dir)                    == str
   assert sim_dir == None or type(sim_dir) == str
   assert type(option_predict)             == dict
   assert type(all_node_database)          == str
   assert type(all_covariate_table)        == list
   assert type( all_covariate_table[0] )   == dict
   assert type(job_table)                  == list
   assert type(node_table)                 == list
   assert type(node_table[0])              == dict
   assert type(root_node_id)               == int
   assert type(root_split_reference_id)    == int
   assert type(at_cascade_log_dict)        == dict
   assert type(job_status_name)            == list
   assert type( job_status_name[0] )       == str
   assert type(shared_job_status_name)     == str
   assert type(shared_lock)                == multiprocessing.synchronize.Lock
   # END_DEF
   # ----------------------------------------------------------------------
   job_status_skip  = job_status_name.index( 'skip' )
   job_status_ready = job_status_name.index( 'ready' )
   job_status_run   = job_status_name.index( 'run' )
   job_status_done  = job_status_name.index( 'done' )
   # ----------------------------------------------------------------------
   #
   # shm_job_status, shared_job_status
   tmp    = numpy.empty(len(job_table), dtype = int )
   mapped = at_cascade.map_shared( shared_job_status_name )
   shm_job_status = multiprocessing.shared_memory.SharedMemory(
      create = False, size = tmp.nbytes, name = mapped
   )
   shared_job_status = numpy.ndarray(
      tmp.shape, dtype = tmp.dtype, buffer = shm_job_status.buf
   )
   #
   # job_table_index
   job_table_index = numpy.array( range(len(job_table)), dtype = int)
   #
   # split_reference_table
   split_reference_table = at_cascade.csv.split_reference_table
   #
   # float_precision
   float_precision = option_predict['float_precision']
   #
   # n_skip
   n_skip = None
   #
   while True :
      #
      # Begin Lock
      acquire_lock(shared_lock)
      #
      # n_skip, predict_job_id, shared_job_status
      #
      job_id_ready = job_table_index[shared_job_status == job_status_ready]
      job_id_skip  = job_table_index[shared_job_status == job_status_skip]
      #
      # if n_ready is zero, all jobs are done or running
      n_ready      = len(job_id_ready)
      if n_ready == 0 :
         shared_lock.release()
         shm_job_status.close()
         return
      #
      if n_skip == None :
         n_skip = len(job_id_skip)
      assert n_skip == len(job_id_skip)
      #
      predict_job_id                    = int( job_id_ready[0] )
      shared_job_status[predict_job_id] = job_status_run
      #
      # End Lock
      shared_lock.release()
      # -------------------------------------------------------------------
      #
      # predict_job_name, predict_node_id, predict_sex_id
      predict_job_row         = job_table[predict_job_id]
      predict_job_name        = predict_job_row['job_name']
      predict_node_id         = predict_job_row['fit_node_id']
      predict_sex_id          = predict_job_row['split_reference_id']
      #
      # print_begin
      print_begin(job_name = predict_job_name)
      #
      # predict_job_dir, ancestor_job_dir
      predict_job_dir, ancestor_job_dir = at_cascade.csv.ancestor_fit(
         fit_dir                 = fit_dir ,
         job_table               = job_table ,
         predict_job_id          = predict_job_id ,
         node_table              = node_table ,
         root_node_id            = root_node_id ,
         split_reference_table   = split_reference_table ,
         root_split_reference_id = root_split_reference_id ,
         at_cascade_log_dict     = at_cascade_log_dict ,
         allow_same_job          = True ,
      )
      if ancestor_job_dir == None :
         if predict_job_id == 0 :
            msg = f'Cannot find a fit for the root job {predict_job_name}'
         else :
            msg = f'Cannot find a fit for {predict_job_name}'
            msg += ' or any of its ancestors'
            assert False, msg
      #
      # predict_directory
      predict_directory = f'{fit_dir}/{predict_job_dir}'
      os.makedirs( f'{predict_directory}', exist_ok = True )
      for prefix in [ 'fit', 'sam', 'tru' ] :
         for suffix in [ 'prior', 'posterior' ] :
            output_file = f'{predict_directory}/{prefix}_{suffix}.csv'
            if os.path.exists( output_file ) :
               os.remove( output_file )
      #
      # predict_job_error
      predict_job_error = None
      if ancestor_job_dir == predict_job_dir :
         #
         # fit_same_as_predict, pre_database
         fit_same_as_predict = True
         fit_database        = f'{predict_directory}/dismod.db'
         pre_database        = f'{predict_directory}/this.db'
         shutil.copyfile(fit_database, pre_database)
         #
         # try_one_job, predict_job_error
         predict_job_error   = try_one_job(
            predict_job_name        = predict_job_name          ,
            fit_dir                 = fit_dir                   ,
            sim_dir                 = sim_dir                   ,
            pre_database            = pre_database              ,
            predict_node_id         = predict_node_id           ,
            predict_sex_id          = predict_sex_id            ,
            all_node_database       = all_node_database         ,
            all_covariate_table     = all_covariate_table       ,
            float_precision         = float_precision           ,
            fit_same_as_predict     = fit_same_as_predict       ,
            option_predict          = option_predict            ,
         )
         #
         # predict_job_dir, ancestor_job_dir
         predict_job_dir, ancestor_job_dir = at_cascade.csv.ancestor_fit(
            fit_dir                 = fit_dir ,
            job_table               = job_table ,
            predict_job_id          = predict_job_id ,
            node_table              = node_table ,
            root_node_id            = root_node_id ,
            split_reference_table   = split_reference_table ,
            root_split_reference_id = root_split_reference_id ,
            at_cascade_log_dict     = at_cascade_log_dict ,
            allow_same_job          = False ,
         )
         assert predict_job_dir != ancestor_job_dir
      #
      if ancestor_job_dir == None :
         assert predict_job_id == 0
      else :
         assert predict_job_dir != ancestor_job_dir
         #
         # fit_same_as_predict, pre_database
         fit_same_as_predict = False
         fit_database   = f'{fit_dir}/{ancestor_job_dir}/dismod.db'
         pre_database   = f'{predict_directory}/ancestor.db'
         shutil.copyfile(fit_database, pre_database)
         #
         # pre_database
         level             = predict_job_dir.count('/') + 1
         path2root_node_db = level * '../' + 'root.db'
         command = [
            'dismod_at', pre_database,
            'set', 'option', 'other_database', path2root_node_db
         ]
         dismod_at.system_command_prc(command, print_command = False)
         #
         # try_one_job, prior_job_error
         prior_job_error = try_one_job(
            predict_job_name        = predict_job_name          ,
            fit_dir                 = fit_dir                   ,
            sim_dir                 = sim_dir                   ,
            pre_database            = pre_database              ,
            predict_node_id         = predict_node_id           ,
            predict_sex_id          = predict_sex_id            ,
            all_node_database       = all_node_database         ,
            all_covariate_table     = all_covariate_table       ,
            float_precision         = float_precision           ,
            fit_same_as_predict     = fit_same_as_predict       ,
            option_predict          = option_predict            ,
         )
         if predict_job_error == None :
            predict_job_error = prior_job_error
         elif prior_job_error != None :
            predict_job_error = f'{prior_job_error}: {predict_job_error}'
      #
      # Begin Lock
      acquire_lock(shared_lock)
      #
      # shared_job_status, n_done
      assert shared_job_status[predict_job_id] == job_status_run
      shared_job_status[predict_job_id] = job_status_done
      job_id_done = job_table_index[ shared_job_status == job_status_done]
      n_done      = len(job_id_done)
      n_total     = len(job_table) - n_skip
      #
      # print_end
      print_end(
         job_name    = predict_job_name,
         n_done      = n_done,
         n_job_total = n_total,
         job_error   = predict_job_error,
      )
      #
      # End Lock
      shared_lock.release()
      # -------------------------------------------------------------------



# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: University of Washington <https://www.washington.edu>
# SPDX-FileContributor: 2021-24 Bradley M. Bell
# ----------------------------------------------------------------------------
'''
{xrst_begin empty_avgint_table}

Create An Empty avgint Table
############################

Prototype
*********
{xrst_literal
   # BEGIN_DEF
   # END_DEF
}

connection
**********
is a dismod_at open connection to the database.

covariate Table
***************
There must be a covariate table in the database so
this procedure can determine the number of covariates.

avgint Table
************
Any previous avgint table is lost.

log Table
*********
The message ``empty_avgint_table`` is added at the end of the log table
with message type ``at_cascade``.

{xrst_end   empty_avgint_table}
'''
import time
import dismod_at
import at_cascade
# ----------------------------------------------------------------------------
# BEGIN_DEF
# at_cascade.empty_avgint_table
def empty_avgint_table(connection) :
# END_DEF
# ----------------------------------------------------------------------------
   #
   # n_covariate
   covariate_table = dismod_at.get_table_dict(connection, 'covariate')
   n_covariate     = len(covariate_table)
   #
   # col_name
   col_name = [
      'integrand_id',
      'node_id',
      'subgroup_id',
      'weight_id',
      'age_lower',
      'age_upper',
      'time_lower',
      'time_upper',
   ]
   for covariate_id in range(n_covariate) :
      col_name.append(f'x_{covariate_id}')
   #
   # col_type
   col_type = [
      'integer',
      'integer',
      'integer',
      'integer',
      'real',
      'real',
      'real',
      'real',
   ]
   for covariate_id in range(n_covariate) :
      col_type.append('real')
   #
   # create_table
   row_list = list()
   tbl_name = 'avgint'
   dismod_at.create_table(
      connection, tbl_name, col_name, col_type, row_list
   )
   #
   # add_log_entry
   message = 'empty_avgint_table'
   at_cascade.add_log_entry(connection, message)
